#
# Full model for Ohm's law in the electrode
#
import pybamm

from .base_ohm import BaseModel


class LeadingOrder(BaseModel):
    """An electrode submodel that employs Ohm's law the leading-order approximation to
    governing equations.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        Either 'Negative' or 'Positive'
    options : dict, optional
        A dictionary of options to be passed to the model.
    set_positive_potential :  bool, optional
        If True the battery model sets the positve potential based on the current.
        If False, the potential is specified by the user. Default is True.

    **Extends:** :class:`pybamm.electrode.ohm.BaseModel`
    """

    def __init__(self, param, domain, options=None, set_positive_potential=True):
        super().__init__(
            param,
            domain,
            options=options,
            set_positive_potential=set_positive_potential,
        )

    def get_coupled_variables(self, variables):
        """
        Returns variables which are derived from the fundamental variables in the model.
        """
        param = self.param

        i_boundary_cc = variables["Current collector current density"]
        phi_s_cn = variables["Negative current collector potential"]

        # import parameters and spatial variables
        l_n = param.l_n
        l_p = param.l_p
        x_n = pybamm.standard_spatial_vars.x_n
        x_p = pybamm.standard_spatial_vars.x_p

        if self.domain == "Negative":
            phi_s = pybamm.PrimaryBroadcast(phi_s_cn, "negative electrode")
            i_s = i_boundary_cc * (1 - x_n / l_n)

        elif self.domain == "Positive":
            # recall delta_phi = phi_s - phi_e
            delta_phi_p_av = variables[
                "X-averaged positive electrode surface potential difference"
            ]
            phi_e_p_av = variables["X-averaged positive electrolyte potential"]

            v = delta_phi_p_av + phi_e_p_av

            phi_s = pybamm.PrimaryBroadcast(v, ["positive electrode"])
            i_s = i_boundary_cc * (1 - (1 - x_p) / l_p)

        variables.update(self._get_standard_potential_variables(phi_s))
        variables.update(self._get_standard_current_variables(i_s))

        if self.domain == "Positive":
            variables.update(self._get_standard_whole_cell_variables(variables))

        return variables

    def set_boundary_conditions(self, variables):

        phi_s = variables[self.domain + " electrode potential"]

        lbc = (pybamm.Scalar(0), "Neumann")
        rbc = (pybamm.Scalar(0), "Neumann")

        self.boundary_conditions[phi_s] = {"left": lbc, "right": rbc}



#
# Method for plotting voltage components
#
import numpy as np


def plot_voltage_components(
    solution, ax=None, show_legend=True, testing=False, **kwargs_fill
):
    """
    Generate a plot showing the component overpotentials that make up the voltage

    Parameters
    ----------
    solution : :class:`pybamm.Solution`
        Solution object from which to extract voltage components
    ax : matplotlib Axis, optional
        The axis on which to put the plot. If None, a new figure and axis is created.
    show_legend : bool, optional
        Whether to display the legend. Default is True
    testing : bool, optional
        Whether to actually make the plot (turned off for unit tests)
    kwargs_fill
        Keyword arguments, passed to ax.fill_between

    """
    import matplotlib.pyplot as plt

    # Set a default value for alpha, the opacity
    kwargs_fill = {"alpha": 0.6, **kwargs_fill}

    if ax is not None:
        testing = True
    else:
        _, ax = plt.subplots()

    overpotentials = [
        "X-averaged battery reaction overpotential [V]",
        "X-averaged battery concentration overpotential [V]",
        "X-averaged battery electrolyte ohmic losses [V]",
        "X-averaged battery solid phase ohmic losses [V]",
    ]

    # Plot
    # Initialise
    time = solution["Time [h]"].entries
    initial_ocv = solution["X-averaged battery open circuit voltage [V]"](0)
    ocv = solution["X-averaged battery open circuit voltage [V]"].entries
    ax.fill_between(time, ocv, initial_ocv, **kwargs_fill)
    top = ocv
    # Plot components
    for overpotential in overpotentials:
        bottom = top + solution[overpotential].entries
        ax.fill_between(time, bottom, top, **kwargs_fill)
        top = bottom
    V = solution["Battery voltage [V]"].entries
    ax.plot(time, V, "k--")
    if show_legend:
        labels = [
            "Voltage",
            "Open-circuit voltage",
            "Reaction overpotential",
            "Concentration overpotential",
            "Ohmic electrolyte overpotential",
            "Ohmic electrode overpotential",
        ]
        leg = ax.legend(labels, loc="lower left", frameon=True)
        leg.get_frame().set_edgecolor("k")

    # Labels
    ax.set_xlim([time[0], time[-1]])
    ax.set_xlabel("Time [h]")

    y_min, y_max = 0.98 * np.nanmin(V), 1.02 * np.nanmax(initial_ocv)
    ax.set_ylim([y_min, y_max])

    if not testing:  # pragma: no cover
        plt.show()

    return ax



#
# Class for composite surface form electrolyte conductivity employing stefan-maxwell
#
import pybamm

from ..composite_conductivity import Composite


class BaseModel(Composite):
    """
    Base class for composite conservation of charge in the electrolyte employing
    the Stefan-Maxwell constitutive equations employing the surface potential difference
    formulation.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain in which the model holds


    **Extends:** :class:`pybamm.electrolyte_conductivity.Composite`
    """

    def __init__(self, param, domain):
        super().__init__(param, domain)

    def get_fundamental_variables(self):

        if self.domain == "Negative":
            delta_phi_av = pybamm.standard_variables.delta_phi_n_av
        elif self.domain == "Separator":
            return {}
        elif self.domain == "Positive":
            delta_phi_av = pybamm.standard_variables.delta_phi_p_av

        variables = self._get_standard_average_surface_potential_difference_variables(
            delta_phi_av
        )
        return variables

    def get_coupled_variables(self, variables):
        # Only update coupled variables once
        if self.domain == "Negative":
            variables.update(super().get_coupled_variables(variables))

        phi_s = variables[self.domain + " electrode potential"]
        phi_e = variables[self.domain + " electrolyte potential"]
        delta_phi = phi_s - phi_e
        variables.update(
            self._get_standard_surface_potential_difference_variables(delta_phi)
        )
        return variables

    def set_initial_conditions(self, variables):

        delta_phi = variables[
            "X-averaged "
            + self.domain.lower()
            + " electrode surface potential difference"
        ]
        if self.domain == "Negative":
            delta_phi_init = self.param.U_n(self.param.c_n_init(0), self.param.T_init)
        elif self.domain == "Positive":
            delta_phi_init = self.param.U_p(self.param.c_p_init(1), self.param.T_init)

        self.initial_conditions = {delta_phi: delta_phi_init}

    def set_boundary_conditions(self, variables):
        if self.domain == "Negative":
            phi_e = variables["Electrolyte potential"]
            self.boundary_conditions = {
                phi_e: {
                    "left": (pybamm.Scalar(0), "Neumann"),
                    "right": (pybamm.Scalar(0), "Neumann"),
                }
            }


class CompositeDifferential(BaseModel):
    """
    Composite model for conservation of charge in the electrolyte employing the
    Stefan-Maxwell constitutive equations employing the surface potential difference
    formulation and where capacitance is present.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel


    **Extends:** :class:`BaseModel`
    """

    def __init__(self, param, domain):
        super().__init__(param, domain)

    def set_rhs(self, variables):
        param = self.param

        sum_j = variables[
            "Sum of x-averaged "
            + self.domain.lower()
            + " electrode interfacial current densities"
        ]

        sum_j_av = variables[
            "X-averaged "
            + self.domain.lower()
            + " electrode total interfacial current density"
        ]
        delta_phi = variables[
            "X-averaged "
            + self.domain.lower()
            + " electrode surface potential difference"
        ]

        if self.domain == "Negative":
            C_dl = param.C_dl_n
        elif self.domain == "Positive":
            C_dl = param.C_dl_p

        self.rhs[delta_phi] = 1 / C_dl * (sum_j_av - sum_j)


class CompositeAlgebraic(BaseModel):
    """
    Composite model for conservation of charge in the electrolyte employing the
    Stefan-Maxwell constitutive equations employing the surface potential difference
    formulation.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel


    **Extends:** :class:`BaseModel`
    """

    def __init__(self, param, domain):
        super().__init__(param, domain)

    def set_algebraic(self, variables):
        sum_j = variables[
            "Sum of x-averaged "
            + self.domain.lower()
            + " electrode interfacial current densities"
        ]

        sum_j_av = variables[
            "X-averaged "
            + self.domain.lower()
            + " electrode total interfacial current density"
        ]
        delta_phi = variables[
            "X-averaged "
            + self.domain.lower()
            + " electrode surface potential difference"
        ]

        self.algebraic[delta_phi] = sum_j_av - sum_j



from pybamm import exp


def graphite_ocp_Ramadass2004(sto):
    """
    Graphite Open Circuit Potential (OCP) as a function of the
    stochiometry (theta?). The fit is taken from Ramadass 2004.

    References
    ----------
    .. [1] P. Ramadass, Bala Haran, Parthasarathy M. Gomadam, Ralph White, and Branko
    N. Popov. "Development of First Principles Capacity Fade Model for Li-Ion Cells."
    (2004)
    """

    u_eq = (
        0.7222
        + 0.1387 * sto
        + 0.029 * (sto ** 0.5)
        - 0.0172 / sto
        + 0.0019 / (sto ** 1.5)
        + 0.2808 * exp(0.9 - 15 * sto)
        - 0.7984 * exp(0.4465 * sto - 0.4108)
    )

    return u_eq



#
# Class for explicit surface form potentials
#
import pybamm
from ..base_electrolyte_conductivity import BaseElectrolyteConductivity


class Explicit(BaseElectrolyteConductivity):
    """Class for deriving surface potential difference variables from the electrode
    and electrolyte potentials

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain in which the model holds
    options : dict, optional
        A dictionary of options to be passed to the model.


    **Extends:** :class:`pybamm.electrolyte_conductivity.BaseElectrolyteConductivity`
    """

    def __init__(self, param, domain, options=None):
        super().__init__(param, domain, options)

    def get_coupled_variables(self, variables):
        # skip for separator
        if self.domain == "Separator":
            return variables

        if self.half_cell and self.domain == "Negative":
            domain = "Lithium metal interface"
        else:
            domain = self.domain
        phi_s = variables[domain + " electrode potential"]
        phi_e = variables[domain + " electrolyte potential"]
        delta_phi = phi_s - phi_e
        variables.update(
            self._get_standard_surface_potential_difference_variables(delta_phi)
        )
        if (
            "X-averaged"
            + self.domain.lower()
            + "electrode surface potential difference"
            not in variables
        ):
            delta_phi_av = pybamm.x_average(delta_phi)
            variables.update(
                self._get_standard_average_surface_potential_difference_variables(
                    delta_phi_av
                )
            )
        return variables

    def set_boundary_conditions(self, variables):
        # don't set any boundary conditions
        return



#
# Unary operator classes and methods
#
import numbers

import numpy as np
import sympy
from scipy.sparse import csr_matrix, issparse
from sympy.vector.operators import Divergence as sympy_Divergence
from sympy.vector.operators import Gradient as sympy_Gradient
import pybamm


class UnaryOperator(pybamm.Symbol):
    """
    A node in the expression tree representing a unary operator
    (e.g. '-', grad, div)

    Derived classes will specify the particular operator

    **Extends:** :class:`Symbol`

    Parameters
    ----------
    name : str
        name of the node
    child : :class:`Symbol`
        child node
    """

    def __init__(self, name, child, domain=None, auxiliary_domains=None):
        if isinstance(child, numbers.Number):
            child = pybamm.Scalar(child)
        if domain is None:
            domain = child.domain
        if auxiliary_domains is None:
            auxiliary_domains = child.auxiliary_domains
        super().__init__(
            name, children=[child], domain=domain, auxiliary_domains=auxiliary_domains
        )
        self.child = self.children[0]

    def __str__(self):
        """See :meth:`pybamm.Symbol.__str__()`."""
        return "{}({!s})".format(self.name, self.child)

    def create_copy(self):
        """See :meth:`pybamm.Symbol.new_copy()`."""
        new_child = self.child.new_copy()
        return self._unary_new_copy(new_child)

    def _unary_new_copy(self, child):
        """Make a new copy of the unary operator, with child `child`"""

        return self.__class__(child)

    def _unary_jac(self, child_jac):
        """Calculate the jacobian of a unary operator."""
        raise NotImplementedError

    def _unary_evaluate(self, child):
        """Perform unary operation on a child."""
        raise NotImplementedError(
            f"{self.__class__} does not implement _unary_evaluate."
        )

    def evaluate(self, t=None, y=None, y_dot=None, inputs=None, known_evals=None):
        """See :meth:`pybamm.Symbol.evaluate()`."""
        if known_evals is not None:
            if self.id not in known_evals:
                child, known_evals = self.child.evaluate(
                    t, y, y_dot, inputs, known_evals
                )
                known_evals[self.id] = self._unary_evaluate(child)
            return known_evals[self.id], known_evals
        else:
            child = self.child.evaluate(t, y, y_dot, inputs)
            return self._unary_evaluate(child)

    def _evaluate_for_shape(self):
        """
        Default behaviour: unary operator has same shape as child
        See :meth:`pybamm.Symbol.evaluate_for_shape()`
        """
        return self.children[0].evaluate_for_shape()

    def _evaluates_on_edges(self, dimension):
        """See :meth:`pybamm.Symbol._evaluates_on_edges()`."""
        return self.child.evaluates_on_edges(dimension)

    def is_constant(self):
        """See :meth:`pybamm.Symbol.is_constant()`."""
        return self.child.is_constant()

    def _sympy_operator(self, child):
        """Apply appropriate SymPy operators."""
        return self._unary_evaluate(child)

    def to_equation(self):
        """Convert the node and its subtree into a SymPy equation."""
        if self.print_name is not None:
            return sympy.Symbol(self.print_name)
        else:
            eq1 = self.child.to_equation()
            return self._sympy_operator(eq1)


class Negate(UnaryOperator):
    """
    A node in the expression tree representing a `-` negation operator.

    **Extends:** :class:`UnaryOperator`
    """

    def __init__(self, child):
        """See :meth:`pybamm.UnaryOperator.__init__()`."""
        super().__init__("-", child)

    def __str__(self):
        """See :meth:`pybamm.Symbol.__str__()`."""
        return "{}{!s}".format(self.name, self.child)

    def _diff(self, variable):
        """See :meth:`pybamm.Symbol._diff()`."""
        return -self.child.diff(variable)

    def _unary_jac(self, child_jac):
        """See :meth:`pybamm.UnaryOperator._unary_jac()`."""
        return -child_jac

    def _unary_evaluate(self, child):
        """See :meth:`UnaryOperator._unary_evaluate()`."""
        return -child

    def _unary_new_copy(self, child):
        """See :meth:`UnaryOperator._unary_new_copy()`."""
        return -child


class AbsoluteValue(UnaryOperator):
    """
    A node in the expression tree representing an `abs` operator.

    **Extends:** :class:`UnaryOperator`
    """

    def __init__(self, child):
        """See :meth:`pybamm.UnaryOperator.__init__()`."""
        super().__init__("abs", child)

    def diff(self, variable):
        """See :meth:`pybamm.Symbol.diff()`."""
        child = self.child.new_copy()
        return sign(child) * child.diff(variable)

    def _unary_jac(self, child_jac):
        """See :meth:`pybamm.UnaryOperator._unary_jac()`."""
        child = self.child.new_copy()
        return sign(child) * child_jac

    def _unary_evaluate(self, child):
        """See :meth:`UnaryOperator._unary_evaluate()`."""
        return np.abs(child)

    def _unary_new_copy(self, child):
        """See :meth:`UnaryOperator._unary_new_copy()`."""
        return abs(child)


class Sign(UnaryOperator):
    """
    A node in the expression tree representing a `sign` operator.

    **Extends:** :class:`UnaryOperator`
    """

    def __init__(self, child):
        """See :meth:`pybamm.UnaryOperator.__init__()`."""
        super().__init__("sign", child)

    def diff(self, variable):
        """See :meth:`pybamm.Symbol.diff()`."""
        return pybamm.Scalar(0)

    def _unary_jac(self, child_jac):
        """See :meth:`pybamm.UnaryOperator._unary_jac()`."""
        return pybamm.Scalar(0)

    def _unary_evaluate(self, child):
        """See :meth:`UnaryOperator._unary_evaluate()`."""
        if issparse(child):
            return csr_matrix.sign(child)
        else:
            with np.errstate(invalid="ignore"):
                return np.sign(child)

    def _unary_new_copy(self, child):
        """See :meth:`UnaryOperator._unary_new_copy()`."""
        return sign(child)

    def _sympy_operator(self, child):
        """Override :meth:`pybamm.UnaryOperator._sympy_operator`"""
        return sympy.functions.elementary.complexes.sign(child)


class Floor(UnaryOperator):
    """
    A node in the expression tree representing an `floor` operator.

    **Extends:** :class:`UnaryOperator`
    """

    def __init__(self, child):
        """See :meth:`pybamm.UnaryOperator.__init__()`."""
        super().__init__("floor", child)

    def diff(self, variable):
        """See :meth:`pybamm.Symbol.diff()`."""
        return pybamm.Scalar(0)

    def _unary_jac(self, child_jac):
        """See :meth:`pybamm.UnaryOperator._unary_jac()`."""
        return pybamm.Scalar(0)

    def _unary_evaluate(self, child):
        """See :meth:`UnaryOperator._unary_evaluate()`."""
        return np.floor(child)


class Ceiling(UnaryOperator):
    """
    A node in the expression tree representing a `ceil` operator.

    **Extends:** :class:`UnaryOperator`
    """

    def __init__(self, child):
        """See :meth:`pybamm.UnaryOperator.__init__()`."""
        super().__init__("ceil", child)

    def diff(self, variable):
        """See :meth:`pybamm.Symbol.diff()`."""
        return pybamm.Scalar(0)

    def _unary_jac(self, child_jac):
        """See :meth:`pybamm.UnaryOperator._unary_jac()`."""
        return pybamm.Scalar(0)

    def _unary_evaluate(self, child):
        """See :meth:`UnaryOperator._unary_evaluate()`."""
        return np.ceil(child)


class Index(UnaryOperator):
    """
    A node in the expression tree, which stores the index that should be
    extracted from its child after the child has been evaluated.

    Parameters
    ----------
    child : :class:`pybamm.Symbol`
        The symbol of which to take the index
    index : int or slice
        The index (if int) or indices (if slice) to extract from the symbol
    name : str, optional
        The name of the symbol
    check_size : bool, optional
        Whether to check if the slice size exceeds the child size. Default is True.
        This should always be True when creating a new symbol so that the appropriate
        check is performed, but should be False for creating a new copy to avoid
        unnecessarily repeating the check.
    """

    def __init__(self, child, index, name=None, check_size=True):
        self.index = index
        if index == -1:
            self.slice = slice(-1, None)
            if name is None:
                name = "Index[-1]"
        elif isinstance(index, int):
            self.slice = slice(index, index + 1)
            if name is None:
                name = "Index[" + str(index) + "]"
        elif isinstance(index, slice):
            self.slice = index
            if name is None:
                if index.start is None:
                    name = "Index[:{:d}]".format(index.stop)
                else:
                    name = "Index[{:d}:{:d}]".format(index.start, index.stop)
        else:
            raise TypeError("index must be integer or slice")

        if check_size:
            if self.slice in (slice(0, 1), slice(-1, None)):
                pass
            elif self.slice.stop > child.size:
                raise ValueError("slice size exceeds child size")

        super().__init__(name, child)

        # no domain for integer value key
        if isinstance(index, int):
            self.clear_domains()

    def _unary_jac(self, child_jac):
        """See :meth:`pybamm.UnaryOperator._unary_jac()`."""

        # if child.jac returns a matrix of zeros, this subsequently gives a bug
        # when trying to simplify the node Index(child_jac). Instead, search the
        # tree for StateVectors and return a matrix of zeros of the correct size
        # if none are found.
        if not self.has_symbol_of_classes(pybamm.StateVector):
            jac = csr_matrix((1, child_jac.shape[1]))
            return pybamm.Matrix(jac)
        else:
            return Index(child_jac, self.index)

    def set_id(self):
        """See :meth:`pybamm.Symbol.set_id()`"""
        self._id = hash(
            (
                self.__class__,
                self.name,
                self.slice.start,
                self.slice.stop,
                self.children[0].id,
            )
            + tuple(self.domain)
        )

    def _unary_evaluate(self, child):
        """See :meth:`UnaryOperator._unary_evaluate()`."""
        return child[self.slice]

    def _unary_new_copy(self, child):
        """See :meth:`UnaryOperator._unary_new_copy()`."""
        new_index = self.__class__(child, self.index, check_size=False)
        # Keep same domains
        new_index.copy_domains(self)
        return new_index

    def _evaluate_for_shape(self):
        return self._unary_evaluate(self.children[0].evaluate_for_shape())

    def _evaluates_on_edges(self, dimension):
        """See :meth:`pybamm.Symbol._evaluates_on_edges()`."""
        return False


class SpatialOperator(UnaryOperator):
    """
    A node in the expression tree representing a unary spatial operator
    (e.g. grad, div)

    Derived classes will specify the particular operator

    This type of node will be replaced by the :class:`Discretisation`
    class with a :class:`Matrix`

    **Extends:** :class:`UnaryOperator`

    Parameters
    ----------

    name : str
        name of the node
    child : :class:`Symbol`
        child node
    """

    def __init__(self, name, child, domain=None, auxiliary_domains=None):
        super().__init__(name, child, domain, auxiliary_domains)

    def diff(self, variable):
        """See :meth:`pybamm.Symbol.diff()`."""
        # We shouldn't need this
        raise NotImplementedError


class Gradient(SpatialOperator):
    """
    A node in the expression tree representing a grad operator.

    **Extends:** :class:`SpatialOperator`
    """

    def __init__(self, child):
        if child.domain == []:
            raise pybamm.DomainError(
                "Cannot take gradient of '{}' since its domain is empty. ".format(child)
                + "Try broadcasting the object first, e.g.\n\n"
                "\tpybamm.grad(pybamm.PrimaryBroadcast(symbol, 'domain'))"
            )
        if child.evaluates_on_edges("primary") is True:
            raise TypeError(
                "Cannot take gradient of '{}' since it evaluates on edges".format(child)
            )
        super().__init__("grad", child)

    def _evaluates_on_edges(self, dimension):
        """See :meth:`pybamm.Symbol._evaluates_on_edges()`."""
        return True

    def _unary_new_copy(self, child):
        """See :meth:`UnaryOperator._unary_new_copy()`."""
        return grad(child)

    def _sympy_operator(self, child):
        """Override :meth:`pybamm.UnaryOperator._sympy_operator`"""
        return sympy_Gradient(child)


class Divergence(SpatialOperator):
    """
    A node in the expression tree representing a div operator.

    **Extends:** :class:`SpatialOperator`
    """

    def __init__(self, child):
        if child.domain == []:
            raise pybamm.DomainError(
                "Cannot take divergence of '{}' since its domain is empty. ".format(
                    child
                )
                + "Try broadcasting the object first, e.g.\n\n"
                "\tpybamm.div(pybamm.PrimaryBroadcast(symbol, 'domain'))"
            )
        if child.evaluates_on_edges("primary") is False:
            raise TypeError(
                "Cannot take divergence of '{}' since it does not ".format(child)
                + "evaluate on edges. Usually, a gradient should be taken before the "
                "divergence."
            )
        super().__init__("div", child)

    def _evaluates_on_edges(self, dimension):
        """See :meth:`pybamm.Symbol._evaluates_on_edges()`."""
        return False

    def _unary_new_copy(self, child):
        """See :meth:`UnaryOperator._unary_new_copy()`."""
        return div(child)

    def _sympy_operator(self, child):
        """Override :meth:`pybamm.UnaryOperator._sympy_operator`"""
        return sympy_Divergence(child)


class Laplacian(SpatialOperator):
    """
    A node in the expression tree representing a laplacian operator. This is
    currently only implemeted in the weak form for finite element formulations.

    **Extends:** :class:`SpatialOperator`
    """

    def __init__(self, child):
        super().__init__("laplacian", child)

    def _evaluates_on_edges(self, dimension):
        """See :meth:`pybamm.Symbol._evaluates_on_edges()`."""
        return False


class GradientSquared(SpatialOperator):
    """
    A node in the expression tree representing a the inner product of the grad
    operator with itself. In particular, this is useful in the finite element
    formualtion where we only require the (sclar valued) square of the gradient,
    and  not the gradient itself.
    **Extends:** :class:`SpatialOperator`
    """

    def __init__(self, child):
        super().__init__("grad squared", child)

    def _evaluates_on_edges(self, dimension):
        """See :meth:`pybamm.Symbol._evaluates_on_edges()`."""
        return False


class Mass(SpatialOperator):
    """
    Returns the mass matrix for a given symbol, accounting for Dirchlet boundary
    conditions where necessary (e.g. in the finite element formualtion)
    **Extends:** :class:`SpatialOperator`
    """

    def __init__(self, child):
        super().__init__("mass", child)

    def _evaluate_for_shape(self):
        return pybamm.evaluate_for_shape_using_domain(self.domain, typ="matrix")


class BoundaryMass(SpatialOperator):
    """
    Returns the mass matrix for a given symbol assembled over the boundary of
    the domain, accounting for Dirchlet boundary conditions where necessary
    (e.g. in the finite element formualtion)
    **Extends:** :class:`SpatialOperator`
    """

    def __init__(self, child):
        super().__init__("boundary mass", child)

    def _evaluate_for_shape(self):
        return pybamm.evaluate_for_shape_using_domain(self.domain, typ="matrix")


class Integral(SpatialOperator):
    """
    A node in the expression tree representing an integral operator.

    .. math::
        I = \\int_{a}^{b}\\!f(u)\\,du,

    where :math:`a` and :math:`b` are the left-hand and right-hand boundaries of
    the domain respectively, and :math:`u\\in\\text{domain}`.

    Parameters
    ----------
    function : :class:`pybamm.Symbol`
        The function to be integrated (will become self.children[0])
    integration_variable : :class:`pybamm.IndependentVariable`
        The variable over which to integrate

    **Extends:** :class:`SpatialOperator`
    """

    def __init__(self, child, integration_variable):
        if not isinstance(integration_variable, list):
            integration_variable = [integration_variable]

        name = "integral"
        for var in integration_variable:
            if isinstance(var, pybamm.SpatialVariable):
                # Check that child and integration_variable domains agree
                if var.domain == child.domain:
                    self._integration_dimension = "primary"
                elif (
                    "secondary" in child.auxiliary_domains
                    and var.domain == child.auxiliary_domains["secondary"]
                ):
                    self._integration_dimension = "secondary"
                elif (
                    "tertiary" in child.auxiliary_domains
                    and var.domain == child.auxiliary_domains["tertiary"]
                ):
                    self._integration_dimension = "tertiary"
                elif (
                    "quaternary" in child.auxiliary_domains
                    and var.domain == child.auxiliary_domains["quaternary"]
                ):
                    self._integration_dimension = "quaternary"
                else:
                    raise pybamm.DomainError(
                        "integration_variable must be the same as child domain or "
                        "an auxiliary domain"
                    )
            else:
                raise TypeError(
                    "integration_variable must be of type pybamm.SpatialVariable, "
                    "not {}".format(type(var))
                )
            name += " d{}".format(var.name)

        if self._integration_dimension == "primary":
            # integral of a child takes the domain from auxiliary domain of the child
            if child.auxiliary_domains != {}:
                domain = child.auxiliary_domains["secondary"]
                if "tertiary" in child.auxiliary_domains:
                    auxiliary_domains = {
                        "secondary": child.auxiliary_domains["tertiary"]
                    }
                    if "quaternary" in child.auxiliary_domains:
                        auxiliary_domains["tertiary"] = child.auxiliary_domains[
                            "quaternary"
                        ]
                else:
                    auxiliary_domains = {}
            # if child has no auxiliary domain, integral removes domain
            else:
                domain = []
                auxiliary_domains = {}
        elif self._integration_dimension == "secondary":
            # integral in the secondary dimension keeps the same domain, moves
            # quaternary to tertiary and tertiary to secondary domain
            domain = child.domain
            if "tertiary" in child.auxiliary_domains:
                auxiliary_domains = {"secondary": child.auxiliary_domains["tertiary"]}
                if "quaternary" in child.auxiliary_domains:
                    auxiliary_domains["tertiary"] = child.auxiliary_domains[
                        "quaternary"
                    ]
            else:
                auxiliary_domains = {}
        elif self._integration_dimension == "tertiary":
            # integral in the tertiary dimension keeps the domain and secondary domain,
            # moves quaternary to tertiary
            domain = child.domain
            auxiliary_domains = {"secondary": child.auxiliary_domains["secondary"]}
            if "quaternary" in child.auxiliary_domains:
                auxiliary_domains["tertiary"] = child.auxiliary_domains["quaternary"]
        elif self._integration_dimension == "quaternary":
            # integral in the quaternary dimension keeps the domain, secondary and
            # tertiary domains
            domain = child.domain
            auxiliary_domains = {
                "secondary": child.auxiliary_domains["secondary"],
                "tertiary": child.auxiliary_domains["tertiary"],
            }
        if any(isinstance(var, pybamm.SpatialVariable) for var in integration_variable):
            name += " {}".format(child.domain)

        self._integration_variable = integration_variable
        super().__init__(
            name, child, domain=domain, auxiliary_domains=auxiliary_domains
        )

    @property
    def integration_variable(self):
        return self._integration_variable

    def set_id(self):
        """See :meth:`pybamm.Symbol.set_id()`"""
        self._id = hash(
            (self.__class__, self.name)
            + tuple(
                [
                    integration_variable.id
                    for integration_variable in self.integration_variable
                ]
            )
            + (self.children[0].id,)
            + tuple(self.domain)
        )

    def _unary_new_copy(self, child):
        """See :meth:`UnaryOperator._unary_new_copy()`."""

        return self.__class__(child, self.integration_variable)

    def _evaluate_for_shape(self):
        """See :meth:`pybamm.Symbol.evaluate_for_shape_using_domain()`"""
        return pybamm.evaluate_for_shape_using_domain(
            self.domain, self.auxiliary_domains
        )

    def _evaluates_on_edges(self, dimension):
        """See :meth:`pybamm.Symbol._evaluates_on_edges()`."""
        return False

    def _sympy_operator(self, child):
        """Override :meth:`pybamm.UnaryOperator._sympy_operator`"""
        return sympy.Integral(child, sympy.Symbol("xn"))


class BaseIndefiniteIntegral(Integral):
    """
    Base class for indefinite integrals (forward or backward).

    Parameters
    ----------
    function : :class:`pybamm.Symbol`
        The function to be integrated (will become self.children[0])
    integration_variable : :class:`pybamm.IndependentVariable`
        The variable over which to integrate

    **Extends:** :class:`Integral`
    """

    def __init__(self, child, integration_variable):
        if isinstance(integration_variable, list):
            if len(integration_variable) > 1:
                raise NotImplementedError(
                    "Indefinite integral only implemeted w.r.t. one variable"
                )
            else:
                integration_variable = integration_variable[0]
        super().__init__(child, integration_variable)
        # overwrite domains with child domains
        self.copy_domains(child)

    def _evaluate_for_shape(self):
        return self.children[0].evaluate_for_shape()

    def _evaluates_on_edges(self, dimension):
        # If child evaluates on edges, indefinite integral doesn't
        # If child doesn't evaluate on edges, indefinite integral does
        return not self.child.evaluates_on_edges(dimension)


class IndefiniteIntegral(BaseIndefiniteIntegral):
    """
    A node in the expression tree representing an indefinite integral operator.

    .. math::
        I = \\int_{x_\text{min}}^{x}\\!f(u)\\,du

    where :math:`u\\in\\text{domain}` which can represent either a
    spatial or temporal variable.

    Parameters
    ----------
    function : :class:`pybamm.Symbol`
        The function to be integrated (will become self.children[0])
    integration_variable : :class:`pybamm.IndependentVariable`
        The variable over which to integrate

    **Extends:** :class:`BaseIndefiniteIntegral`
    """

    def __init__(self, child, integration_variable):
        super().__init__(child, integration_variable)
        # Overwrite the name
        self.name = "{} integrated w.r.t {}".format(
            child.name, self.integration_variable[0].name
        )
        if isinstance(integration_variable, pybamm.SpatialVariable):
            self.name += " on {}".format(self.integration_variable[0].domain)


class BackwardIndefiniteIntegral(BaseIndefiniteIntegral):
    """
    A node in the expression tree representing a backward indefinite integral operator.

    .. math::
        I = \\int_{x}^{x_\text{max}}\\!f(u)\\,du

    where :math:`u\\in\\text{domain}` which can represent either a
    spatial or temporal variable.

    Parameters
    ----------
    function : :class:`pybamm.Symbol`
        The function to be integrated (will become self.children[0])
    integration_variable : :class:`pybamm.IndependentVariable`
        The variable over which to integrate

    **Extends:** :class:`BaseIndefiniteIntegral`
    """

    def __init__(self, child, integration_variable):
        super().__init__(child, integration_variable)
        # Overwrite the name
        self.name = "{} integrated backward w.r.t {}".format(
            child.name, self.integration_variable[0].name
        )
        if isinstance(integration_variable, pybamm.SpatialVariable):
            self.name += " on {}".format(self.integration_variable[0].domain)


class DefiniteIntegralVector(SpatialOperator):
    """
    A node in the expression tree representing an integral of the basis used
    for discretisation

    .. math::
        I = \\int_{a}^{b}\\!\\psi(x)\\,dx,

    where :math:`a` and :math:`b` are the left-hand and right-hand boundaries of
    the domain respectively and :math:`\\psi` is the basis function.

    Parameters
    ----------
    variable : :class:`pybamm.Symbol`
        The variable whose basis will be integrated over the entire domain
    vector_type : str, optional
        Whether to return a row or column vector (default is row)

    **Extends:** :class:`SpatialOperator`
    """

    def __init__(self, child, vector_type="row"):
        name = "basis integral"
        self.vector_type = vector_type
        super().__init__(name, child)
        # integrating removes the domain
        self.clear_domains()

    def set_id(self):
        """See :meth:`pybamm.Symbol.set_id()`"""
        self._id = hash(
            (self.__class__, self.name, self.vector_type)
            + (self.children[0].id,)
            + tuple(self.domain)
        )

    def _unary_new_copy(self, child):
        """See :meth:`UnaryOperator._unary_new_copy()`."""

        return self.__class__(child, vector_type=self.vector_type)

    def _evaluate_for_shape(self):
        """See :meth:`pybamm.Symbol.evaluate_for_shape_using_domain()`"""
        return pybamm.evaluate_for_shape_using_domain(self.domain)


class BoundaryIntegral(SpatialOperator):
    """
    A node in the expression tree representing an integral operator over the
    boundary of a domain

    .. math::
        I = \\int_{\\partial a}\\!f(u)\\,du,

    where :math:`\\partial a` is the boundary of the domain, and
    :math:`u\\in\\text{domain boundary}`.

    Parameters
    ----------
    function : :class:`pybamm.Symbol`
        The function to be integrated (will become self.children[0])
    region : str, optional
        The region of the boundary over which to integrate. If region is `entire`
        (default) the integration is carried out over the entire boundary. If
        region is `negative tab` or `positive tab` then the integration is only
        carried out over the appropriate part of the boundary corresponding to
        the tab.
    **Extends:** :class:`SpatialOperator`
    """

    def __init__(self, child, region="entire"):
        # boundary integral removes domain
        domain = []
        auxiliary_domains = {}

        name = "boundary integral over "
        if region == "entire":
            name += "entire boundary"
        elif region == "negative tab":
            name += "negative tab"
        elif region == "positive tab":
            name += "positive tab"
        self.region = region
        super().__init__(
            name, child, domain=domain, auxiliary_domains=auxiliary_domains
        )

    def set_id(self):
        """See :meth:`pybamm.Symbol.set_id()`"""
        self._id = hash(
            (self.__class__, self.name) + (self.children[0].id,) + tuple(self.domain)
        )

    def _unary_new_copy(self, child):
        """See :meth:`UnaryOperator._unary_new_copy()`."""

        return self.__class__(child, region=self.region)

    def _evaluate_for_shape(self):
        """See :meth:`pybamm.Symbol.evaluate_for_shape_using_domain()`"""
        return pybamm.evaluate_for_shape_using_domain(self.domain)

    def _evaluates_on_edges(self, dimension):
        """See :meth:`pybamm.Symbol._evaluates_on_edges()`."""
        return False


class DeltaFunction(SpatialOperator):
    """
    Delta function. Currently can only be implemented at the edge of a domain.

    Parameters
    ----------
    child : :class:`pybamm.Symbol`
        The variable that sets the strength of the delta function
    side : str
        Which side of the domain to implement the delta function on

    **Extends:** :class:`SpatialOperator`
    """

    def __init__(self, child, side, domain):
        self.side = side
        if domain is None:
            raise pybamm.DomainError("Delta function domain cannot be None")
        if child.domain != []:
            auxiliary_domains = {"secondary": child.domain}
        else:
            auxiliary_domains = {}
        super().__init__("delta_function", child, domain, auxiliary_domains)

    def set_id(self):
        """See :meth:`pybamm.Symbol.set_id()`"""
        self._id = hash(
            (self.__class__, self.name, self.side, self.children[0].id)
            + tuple(self.domain)
            + tuple([(k, tuple(v)) for k, v in self.auxiliary_domains.items()])
        )

    def _evaluates_on_edges(self, dimension):
        """See :meth:`pybamm.Symbol._evaluates_on_edges()`."""
        return False

    def _unary_new_copy(self, child):
        """See :meth:`UnaryOperator._unary_new_copy()`."""
        return self.__class__(child, self.side, self.domain)

    def evaluate_for_shape(self):
        """See :meth:`pybamm.Symbol.evaluate_for_shape_using_domain()`"""
        child_eval = self.children[0].evaluate_for_shape()
        vec = pybamm.evaluate_for_shape_using_domain(self.domain)

        return np.outer(child_eval, vec).reshape(-1, 1)


class BoundaryOperator(SpatialOperator):
    """
    A node in the expression tree which gets the boundary value of a variable.

    Parameters
    ----------
    name : str
        The name of the symbol
    child : :class:`pybamm.Symbol`
        The variable whose boundary value to take
    side : str
        Which side to take the boundary value on ("left" or "right")

    **Extends:** :class:`SpatialOperator`
    """

    def __init__(self, name, child, side):
        # side can only be "negative tab" or "positive tab" if domain is
        # "current collector"
        if side in ["negative tab", "positive tab"]:
            if child.domain[0] != "current collector":
                raise pybamm.ModelError(
                    """Can only take boundary value on the tabs in the domain
                'current collector', but {} has domain {}""".format(
                        child, child.domain[0]
                    )
                )
        self.side = side
        # boundary value of a child takes the domain from auxiliary domain of the child
        if child.auxiliary_domains != {}:
            domain = child.auxiliary_domains["secondary"]
        # if child has no auxiliary domain, boundary operator removes domain
        else:
            domain = []
        # shift tertiary and quaternary domains down by one
        if "tertiary" in child.auxiliary_domains:
            auxiliary_domains = {"secondary": child.auxiliary_domains["tertiary"]}
            if "quaternary" in child.auxiliary_domains:
                auxiliary_domains["tertiary"] = child.auxiliary_domains["quaternary"]
        else:
            auxiliary_domains = {}
        super().__init__(
            name, child, domain=domain, auxiliary_domains=auxiliary_domains
        )

    def set_id(self):
        """See :meth:`pybamm.Symbol.set_id()`"""
        self._id = hash(
            (self.__class__, self.name, self.side, self.children[0].id)
            + tuple(self.domain)
            + tuple([(k, tuple(v)) for k, v in self.auxiliary_domains.items()])
        )

    def _unary_new_copy(self, child):
        """See :meth:`UnaryOperator._unary_new_copy()`."""
        return self.__class__(child, self.side)

    def _evaluate_for_shape(self):
        """See :meth:`pybamm.Symbol.evaluate_for_shape_using_domain()`"""
        return pybamm.evaluate_for_shape_using_domain(
            self.domain, self.auxiliary_domains
        )


class BoundaryValue(BoundaryOperator):
    """
    A node in the expression tree which gets the boundary value of a variable.

    Parameters
    ----------
    child : :class:`pybamm.Symbol`
        The variable whose boundary value to take
    side : str
        Which side to take the boundary value on ("left" or "right")

    **Extends:** :class:`BoundaryOperator`
    """

    def __init__(self, child, side):
        super().__init__("boundary value", child, side)

    def _unary_new_copy(self, child):
        """See :meth:`UnaryOperator._unary_new_copy()`."""
        return boundary_value(child, self.side)

    def _sympy_operator(self, child):
        """Override :meth:`pybamm.UnaryOperator._sympy_operator`"""
        if (
            self.child.domain[0] in ["negative particle", "positive particle"]
            and self.side == "right"
        ):
            # value on the surface of the particle
            if str(child) == "1":
                return child
            else:
                latex_child = sympy.latex(child) + r"^{surf}"
                return sympy.Symbol(latex_child)

        elif self.side == "positive tab":
            return child

        else:
            latex_child = sympy.latex(child) + r"^{" + sympy.latex(self.side) + r"}"
            return sympy.Symbol(latex_child)


class BoundaryGradient(BoundaryOperator):
    """
    A node in the expression tree which gets the boundary flux of a variable.

    Parameters
    ----------
    child : :class:`pybamm.Symbol`
        The variable whose boundary flux to take
    side : str
        Which side to take the boundary flux on ("left" or "right")

    **Extends:** :class:`BoundaryOperator`
    """

    def __init__(self, child, side):
        super().__init__("boundary flux", child, side)


class UpwindDownwind(SpatialOperator):
    """
    A node in the expression tree representing an upwinding or downwinding operator.
    Usually to be used for better stability in convection-dominated equations.

    **Extends:** :class:`SpatialOperator`
    """

    def __init__(self, name, child):
        if child.domain == []:
            raise pybamm.DomainError(
                "Cannot upwind '{}' since its domain is empty. ".format(child)
                + "Try broadcasting the object first, e.g.\n\n"
                "\tpybamm.div(pybamm.PrimaryBroadcast(symbol, 'domain'))"
            )
        if child.evaluates_on_edges("primary") is True:
            raise TypeError(
                "Cannot upwind '{}' since it does not ".format(child)
                + "evaluate on nodes."
            )
        super().__init__(name, child)

    def _evaluates_on_edges(self, dimension):
        """See :meth:`pybamm.Symbol._evaluates_on_edges()`."""
        return True


class Upwind(UpwindDownwind):
    """
    Upwinding operator. To be used if flow velocity is positive (left to right).

    **Extends:** :class:`UpwindDownwind`
    """

    def __init__(self, child):
        super().__init__("upwind", child)


class Downwind(UpwindDownwind):
    """
    Downwinding operator. To be used if flow velocity is negative (right to left).

    **Extends:** :class:`UpwindDownwind`
    """

    def __init__(self, child):
        super().__init__("downwind", child)


class NotConstant(UnaryOperator):
    """Special class to wrap a symbol that should not be treated as a constant."""

    def __init__(self, child):
        super().__init__("not_constant", child)

    def _unary_new_copy(self, child):
        """See :meth:`pybamm.Symbol.new_copy()`."""
        return NotConstant(child)

    def _diff(self, variable):
        """See :meth:`pybamm.Symbol._diff()`."""
        return self.child.diff(variable)

    def _unary_jac(self, child_jac):
        """See :meth:`pybamm.UnaryOperator._unary_jac()`."""
        return child_jac

    def _unary_evaluate(self, child):
        """See :meth:`UnaryOperator._unary_evaluate()`."""
        return child

    def is_constant(self):
        """See :meth:`pybamm.Symbol.is_constant()`."""
        # This symbol is not constant
        return False


#
# Methods to call Gradient, Divergence, Laplacian and GradientSquared
#


def grad(symbol):
    """
    convenience function for creating a :class:`Gradient`

    Parameters
    ----------

    symbol : :class:`Symbol`
        the gradient will be performed on this sub-symbol

    Returns
    -------

    :class:`Gradient`
        the gradient of ``symbol``
    """
    # Gradient of a broadcast is zero
    if isinstance(symbol, pybamm.PrimaryBroadcast):
        new_child = pybamm.PrimaryBroadcast(0, symbol.child.domain)
        return pybamm.PrimaryBroadcastToEdges(new_child, symbol.domain)
    elif isinstance(symbol, pybamm.FullBroadcast):
        return pybamm.FullBroadcastToEdges(0, symbol.domain, symbol.auxiliary_domains)
    else:
        return Gradient(symbol)


def div(symbol):
    """
    convenience function for creating a :class:`Divergence`

    Parameters
    ----------

    symbol : :class:`Symbol`
        the divergence will be performed on this sub-symbol

    Returns
    -------

    :class:`Divergence`
        the divergence of ``symbol``
    """
    # Divergence of a broadcast is zero
    if isinstance(symbol, pybamm.PrimaryBroadcastToEdges):
        new_child = pybamm.PrimaryBroadcast(0, symbol.child.domain)
        return pybamm.PrimaryBroadcast(new_child, symbol.domain)
    # Divergence commutes with Negate operator
    if isinstance(symbol, pybamm.Negate):
        return -div(symbol.orphans[0])
    elif isinstance(symbol, (pybamm.Multiplication, pybamm.Division)):
        left, right = symbol.orphans
        if isinstance(left, pybamm.Negate):
            return -div(symbol._binary_new_copy(left.orphans[0], right))
        # elif isinstance(right, pybamm.Negate):
        #     return -div(symbol._binary_new_copy(left, right.orphans[0]))

    # Last resort
    return Divergence(symbol)


def laplacian(symbol):
    """
    convenience function for creating a :class:`Laplacian`

    Parameters
    ----------

    symbol : :class:`Symbol`
        the laplacian will be performed on this sub-symbol

    Returns
    -------

    :class:`Laplacian`
        the laplacian of ``symbol``
    """

    return Laplacian(symbol)


def grad_squared(symbol):
    """
    convenience function for creating a :class:`GradientSquared`

    Parameters
    ----------

    symbol : :class:`Symbol`
        the inner product of the gradient with itself will be performed on this
        sub-symbol

    Returns
    -------

    :class:`GradientSquared`
        inner product of the gradient of ``symbol`` with itself
    """

    return GradientSquared(symbol)


def upwind(symbol):
    """convenience function for creating a :class:`Upwind`"""
    return Upwind(symbol)


def downwind(symbol):
    """convenience function for creating a :class:`Downwind`"""
    return Downwind(symbol)


#
# Method to call SurfaceValue
#


def surf(symbol):
    """
    convenience function for creating a right :class:`BoundaryValue`, usually in the
    spherical geometry.

    Parameters
    ----------

    symbol : :class:`pybamm.Symbol`
        the surface value of this symbol will be returned

    Returns
    -------
    :class:`pybamm.BoundaryValue`
        the surface value of ``symbol``
    """
    return boundary_value(symbol, "right")


def boundary_value(symbol, side):
    """
    convenience function for creating a :class:`pybamm.BoundaryValue`

    Parameters
    ----------
    symbol : `pybamm.Symbol`
        The symbol whose boundary value to take
    side : str
        Which side to take the boundary value on ("left" or "right")

    Returns
    -------
    :class:`BoundaryValue`
        the new integrated expression tree
    """
    # Can't take boundary value if the symbol evaluates on edges
    if symbol.evaluates_on_edges("primary"):
        raise ValueError(
            "Can't take the boundary value of a symbol that evaluates on edges"
        )

    # If symbol doesn't have a domain, its boundary value is itself
    if symbol.domain == []:
        new_symbol = symbol.new_copy()
        new_symbol.parent = None
        return new_symbol
    # If symbol is a primary or full broadcast, reduce by one dimension
    if isinstance(symbol, (pybamm.PrimaryBroadcast, pybamm.FullBroadcast)):
        return symbol.reduce_one_dimension()
    # If symbol is a secondary broadcast, its boundary value is a primary broadcast of
    # the boundary value of its child
    if isinstance(symbol, pybamm.SecondaryBroadcast):
        # Read child (making copy)
        child = symbol.orphans[0]
        # Take boundary value
        boundary_child = boundary_value(child, side)
        # Broadcast back to the original symbol's secondary domain
        return pybamm.PrimaryBroadcast(boundary_child, symbol.secondary_domain)
    # Otherwise, calculate boundary value
    else:
        return BoundaryValue(symbol, side)


def sign(symbol):
    """Returns a :class:`Sign` object."""
    return pybamm.simplify_if_constant(Sign(symbol))


def smooth_absolute_value(symbol, k):
    """
    Smooth approximation to the absolute value function. k is the smoothing parameter,
    set by `pybamm.settings.abs_smoothing`. The recommended value is k=10.
    """
    x = symbol
    exp = pybamm.exp
    kx = k * symbol
    return x * (exp(kx) - exp(-kx)) / (exp(kx) + exp(-kx))



#
# Standard thermal parameters
#
import pybamm
from .base_parameters import BaseParameters


class ThermalParameters(BaseParameters):
    """
    Standard thermal parameters

    Layout:
        1. Dimensional Parameters
        2. Dimensional Functions
        3. Dimensionless Parameters
        4. Dimensionless Functions
    """

    def __init__(self):

        # Get geometric parameters
        self.geo = pybamm.geometric_parameters

        # Set parameters
        self._set_dimensional_parameters()
        self._set_dimensionless_parameters()

    def _set_dimensional_parameters(self):
        """Defines the dimensional parameters"""

        # Reference temperature
        self.T_ref = pybamm.Parameter("Reference temperature [K]")

        # Cooling coefficient
        self.h_cn_dim = pybamm.Parameter(
            "Negative current collector surface heat transfer coefficient [W.m-2.K-1]"
        )
        self.h_cp_dim = pybamm.Parameter(
            "Positive current collector surface heat transfer coefficient [W.m-2.K-1]"
        )
        self.h_tab_n_dim = pybamm.Parameter(
            "Negative tab heat transfer coefficient [W.m-2.K-1]"
        )
        self.h_tab_p_dim = pybamm.Parameter(
            "Positive tab heat transfer coefficient [W.m-2.K-1]"
        )
        self.h_edge_dim = pybamm.Parameter("Edge heat transfer coefficient [W.m-2.K-1]")
        self.h_total_dim = pybamm.Parameter(
            "Total heat transfer coefficient [W.m-2.K-1]"
        )

        # Typical temperature rise
        self.Delta_T = pybamm.Scalar(1)

        # Initial temperature
        self.T_init_dim = pybamm.Parameter("Initial temperature [K]")

        # Planar (y,z) thermal diffusion timescale
        self.tau_th_yz = (
            self.rho_eff_dim(self.T_ref)
            * (self.geo.L_z ** 2)
            / self.lambda_eff_dim(self.T_ref)
        )

    def T_amb_dim(self, t):
        """Dimensional ambient temperature"""
        return pybamm.FunctionParameter("Ambient temperature [K]", {"Time [s]": t})

    def rho_cn_dim(self, T):
        """Negative current collector density [kg.m-3]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            "Negative current collector density [kg.m-3]", inputs
        )

    def rho_n_dim(self, T):
        """Negative electrode density [kg.m-3]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter("Negative electrode density [kg.m-3]", inputs)

    def rho_s_dim(self, T):
        """Separator density [kg.m-3]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter("Separator density [kg.m-3]", inputs)

    def rho_p_dim(self, T):
        """Positive electrode density [kg.m-3]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter("Positive electrode density [kg.m-3]", inputs)

    def rho_cp_dim(self, T):
        """Positive current collector density [kg.m-3]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            "Positive current collector density [kg.m-3]", inputs
        )

    def rho_eff_dim(self, T):
        """Effective volumetric heat capacity [J.m-3.K-1]"""
        return (
            self.rho_cn_dim(T) * self.c_p_cn_dim(T) * self.geo.L_cn
            + self.rho_n_dim(T) * self.c_p_n_dim(T) * self.geo.L_n
            + self.rho_s_dim(T) * self.c_p_s_dim(T) * self.geo.L_s
            + self.rho_p_dim(T) * self.c_p_p_dim(T) * self.geo.L_p
            + self.rho_cp_dim(T) * self.c_p_cp_dim(T) * self.geo.L_cp
        ) / self.geo.L

    def c_p_cn_dim(self, T):
        """Negative current collector specific heat capacity [J.kg-1.K-1]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            "Negative current collector specific heat capacity [J.kg-1.K-1]", inputs
        )

    def c_p_n_dim(self, T):
        """Negative electrode specific heat capacity [J.kg-1.K-1]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            "Negative electrode specific heat capacity [J.kg-1.K-1]", inputs
        )

    def c_p_s_dim(self, T):
        """Separator specific heat capacity [J.kg-1.K-1]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            "Separator specific heat capacity [J.kg-1.K-1]", inputs
        )

    def c_p_p_dim(self, T):
        """Positive electrode specific heat capacity [J.kg-1.K-1]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            "Positive electrode specific heat capacity [J.kg-1.K-1]", inputs
        )

    def c_p_cp_dim(self, T):
        """Positive current collector specific heat capacity [J.kg-1.K-1]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            "Positive current collector specific heat capacity [J.kg-1.K-1]", inputs
        )

    def lambda_cn_dim(self, T):
        """Negative current collector thermal conductivity [W.m-1.K-1]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            "Negative current collector thermal conductivity [W.m-1.K-1]", inputs
        )

    def lambda_n_dim(self, T):
        """Negative electrode thermal conductivity [W.m-1.K-1]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            "Negative electrode thermal conductivity [W.m-1.K-1]", inputs
        )

    def lambda_s_dim(self, T):
        """Separator thermal conductivity [W.m-1.K-1]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            "Separator thermal conductivity [W.m-1.K-1]", inputs
        )

    def lambda_p_dim(self, T):
        """Positive electrode thermal conductivity [W.m-1.K-1]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            "Positive electrode thermal conductivity [W.m-1.K-1]", inputs
        )

    def lambda_cp_dim(self, T):
        """Positive current collector thermal conductivity [W.m-1.K-1]"""
        inputs = {"Temperature [K]": T}
        return pybamm.FunctionParameter(
            "Positive current collector thermal conductivity [W.m-1.K-1]", inputs
        )

    def lambda_eff_dim(self, T):
        """Effective thermal conductivity [W.m-1.K-1]"""
        return (
            self.lambda_cn_dim(T) * self.geo.L_cn
            + self.lambda_n_dim(T) * self.geo.L_n
            + self.lambda_s_dim(T) * self.geo.L_s
            + self.lambda_p_dim(T) * self.geo.L_p
            + self.lambda_cp_dim(T) * self.geo.L_cp
        ) / self.geo.L

    def _set_dimensionless_parameters(self):
        """Defines the dimensionless parameters"""

        # Relative temperature rise
        self.Theta = self.Delta_T / self.T_ref

        # Cooling coefficient
        self.h_cn = self.h_cn_dim * self.geo.L_x / self.lambda_eff_dim(self.T_ref)
        self.h_cp = self.h_cp_dim * self.geo.L_x / self.lambda_eff_dim(self.T_ref)
        self.h_tab_n = self.h_tab_n_dim * self.geo.L_x / self.lambda_eff_dim(self.T_ref)
        self.h_tab_p = self.h_tab_p_dim * self.geo.L_x / self.lambda_eff_dim(self.T_ref)
        self.h_edge = self.h_edge_dim * self.geo.L_x / self.lambda_eff_dim(self.T_ref)
        self.h_total = self.h_total_dim * self.geo.L_x / self.lambda_eff_dim(self.T_ref)

        # Initial temperature
        self.T_init = (self.T_init_dim - self.T_ref) / self.Delta_T

    def T_amb(self, t):
        """Dimensionless ambient temperature"""
        return (self.T_amb_dim(t) - self.T_ref) / self.Delta_T

    def rho_cn(self, T):
        """Dimensionless negative current collector density"""
        T_dim = self.Delta_T * T + self.T_ref
        return (
            self.rho_cn_dim(T_dim)
            * self.c_p_cn_dim(T_dim)
            / self.rho_eff_dim(self.T_ref)
        )

    def rho_n(self, T):
        """Dimensionless negative electrode density"""
        T_dim = self.Delta_T * T + self.T_ref
        return (
            self.rho_n_dim(T_dim) * self.c_p_n_dim(T_dim) / self.rho_eff_dim(self.T_ref)
        )

    def rho_s(self, T):
        """Dimensionless separator density"""
        T_dim = self.Delta_T * T + self.T_ref
        return (
            self.rho_s_dim(T_dim) * self.c_p_s_dim(T_dim) / self.rho_eff_dim(self.T_ref)
        )

    def rho_p(self, T):
        """Dimensionless positive electrode density"""
        T_dim = self.Delta_T * T + self.T_ref
        return (
            self.rho_p_dim(T_dim) * self.c_p_p_dim(T_dim) / self.rho_eff_dim(self.T_ref)
        )

    def rho_cp(self, T):
        """Dimensionless positive current collector density"""
        T_dim = self.Delta_T * T + self.T_ref
        return (
            self.rho_cp_dim(T_dim)
            * self.c_p_cp_dim(T_dim)
            / self.rho_eff_dim(self.T_ref)
        )

    def lambda_cn(self, T):
        """Dimensionless negative current collector thermal conductivity"""
        T_dim = self.Delta_T * T + self.T_ref
        return self.lambda_cn_dim(T_dim) / self.lambda_eff_dim(self.T_ref)

    def lambda_n(self, T):
        """Dimensionless negative electrode thermal conductivity"""
        T_dim = self.Delta_T * T + self.T_ref
        return self.lambda_n_dim(T_dim) / self.lambda_eff_dim(self.T_ref)

    def lambda_s(self, T):
        """Dimensionless separator thermal conductivity"""
        T_dim = self.Delta_T * T + self.T_ref
        return self.lambda_s_dim(T_dim) / self.lambda_eff_dim(self.T_ref)

    def lambda_p(self, T):
        """Dimensionless positive electrode thermal conductivity"""
        T_dim = self.Delta_T * T + self.T_ref
        return self.lambda_p_dim(T_dim) / self.lambda_eff_dim(self.T_ref)

    def lambda_cp(self, T):
        """Dimensionless positive current collector thermal conductivity"""
        T_dim = self.Delta_T * T + self.T_ref
        return self.lambda_cp_dim(T_dim) / self.lambda_eff_dim(self.T_ref)


thermal_parameters = ThermalParameters()



#
# Parameter sets from papers
#
"""
Parameter sets from papers. The 'citation' entry provides a reference to the appropriate
paper in the file "pybamm/CITATIONS.txt". To see which parameter sets have been used in
your simulation, add the line "pybamm.print_citations()" to your script.

Lead-acid parameter sets
------------------------
    * Sulzer2019 :
       - Valentin Sulzer, S. Jon Chapman, Colin P. Please, David A. Howey, and Charles
         W. Monroe. Faster Lead-Acid Battery Simulations from Porous-Electrode Theory:
         Part I. Physical Model. Journal of The Electrochemical Society,
         166(12):A2363–A2371, 2019. doi:10.1149/2.0301910jes.

Lithium-ion parameter sets
--------------------------
    * Ai2020 :
       - Weilong Ai, Ludwig Kraft, Johannes Sturm, Andreas Jossen, and Billy Wu.
         Electrochemical thermal-mechanical modelling of stress inhomogeneity in
         lithium-ion pouch cells. Journal of The Electrochemical Society, 167(1):013512,
         2019. doi:10.1149/2.0122001JES.
    * Chen2020 :
       - Chang-Hui Chen, Ferran Brosa Planella, Kieran O'Regan, Dominika Gastol, W.
         Dhammika Widanage, and Emma Kendrick. Development of Experimental Techniques
         for Parameterization of Multi-scale Lithium-ion Battery Models. Journal of The
         Electrochemical Society, 167(8):080534, 2020. doi:10.1149/1945-7111/ab9050.
    * Chen2020_plating :
       - Chang-Hui Chen, Ferran Brosa Planella, Kieran O'Regan, Dominika Gastol, W.
         Dhammika Widanage, and Emma Kendrick. Development of Experimental Techniques
         for Parameterization of Multi-scale Lithium-ion Battery Models. Journal of The
         Electrochemical Society, 167(8):080534, 2020. doi:10.1149/1945-7111/ab9050.
    * Ecker2015 :
       - Madeleine Ecker, Stefan Käbitz, Izaro Laresgoiti, and Dirk Uwe Sauer.
         Parameterization of a Physico-Chemical Model of a Lithium-Ion Battery: II.
         Model Validation. Journal of The Electrochemical Society, 162(9):A1849–A1857,
         2015. doi:10.1149/2.0541509jes.
       - Madeleine Ecker, Thi Kim Dung Tran, Philipp Dechent, Stefan Käbitz, Alexander
         Warnecke, and Dirk Uwe Sauer. Parameterization of a Physico-Chemical Model of a
         Lithium-Ion Battery: I. Determination of Parameters. Journal of the
         Electrochemical Society, 162(9):A1836–A1848, 2015. doi:10.1149/2.0551509jes.
       - Alastair Hales, Laura Bravo Diaz, Mohamed Waseem Marzook, Yan Zhao, Yatish
         Patel, and Gregory Offer. The cell cooling coefficient: a standard to define
         heat rejection from lithium-ion batteries. Journal of The Electrochemical
         Society, 166(12):A2383, 2019.
       - Giles Richardson, Ivan Korotkin, Rahifa Ranom, Michael Castle, and Jamie M.
         Foster. Generalised single particle models for high-rate operation of graded
         lithium-ion electrodes: systematic derivation and validation. Electrochimica
         Acta, 339:135862, 2020. doi:10.1016/j.electacta.2020.135862.
       - Yan Zhao, Yatish Patel, Teng Zhang, and Gregory J Offer. Modeling the effects
         of thermal gradients induced by tab and surface cooling on lithium ion cell
         performance. Journal of The Electrochemical Society, 165(13):A3169, 2018.
    * Marquis2019 :
       - Scott G. Marquis, Valentin Sulzer, Robert Timms, Colin P. Please, and S. Jon
         Chapman. An asymptotic derivation of a single particle model with electrolyte.
         Journal of The Electrochemical Society, 166(15):A3693–A3706, 2019.
         doi:10.1149/2.0341915jes.
    * Mohtat2020 :
       - Peyman Mohtat, Suhak Lee, Valentin Sulzer, Jason B. Siegel, and Anna G.
         Stefanopoulou. Differential Expansion and Voltage Model for Li-ion Batteries at
         Practical Charging Rates. Journal of The Electrochemical Society,
         167(11):110561, 2020. doi:10.1149/1945-7111/aba5d1.
    * NCA_Kim2011 :
       - Gi-Heon Kim, Kandler Smith, Kyu-Jin Lee, Shriram Santhanagopalan, and Ahmad
         Pesaran. Multi-domain modeling of lithium-ion batteries encompassing
         multi-physics in varied length scales. Journal of the Electrochemical Society,
         158(8):A955–A969, 2011. doi:10.1149/1.3597614.
    * ORegan2021 :
       - Chang-Hui Chen, Ferran Brosa Planella, Kieran O'Regan, Dominika Gastol, W.
         Dhammika Widanage, and Emma Kendrick. Development of Experimental Techniques
         for Parameterization of Multi-scale Lithium-ion Battery Models. Journal of The
         Electrochemical Society, 167(8):080534, 2020. doi:10.1149/1945-7111/ab9050.
       - Kieran O'Regan, Ferran Brosa Planella, W. Dhammika Widanage, and Emma Kendrick.
         Thermal-electrochemical parametrisation of a lithium-ion battery: mapping Li
         concentration and temperature dependencies. Journal of The Electrochemical
         Society, ():, 2021. doi:.
    * Prada2013 :
       - Chang-Hui Chen, Ferran Brosa Planella, Kieran O'Regan, Dominika Gastol, W.
         Dhammika Widanage, and Emma Kendrick. Development of Experimental Techniques
         for Parameterization of Multi-scale Lithium-ion Battery Models. Journal of The
         Electrochemical Society, 167(8):080534, 2020. doi:10.1149/1945-7111/ab9050.
       - Michael J. Lain, James Brandon, and Emma Kendrick. Design strategies for high
         power vs. high energy lithium ion cells. Batteries, 5(4):64, 2019.
         doi:10.3390/batteries5040064.
       - Eric Prada, D. Di Domenico, Y. Creff, J. Bernard, Valérie Sauvant-Moynot, and
         François Huet. A simplified electrochemical and thermal aging model of
         LiFePO4-graphite Li-ion batteries: power and capacity fade simulations. Journal
         of The Electrochemical Society, 160(4):A616, 2013. doi:10.1149/2.053304jes.
    * Ramadass2004 :
       - P Ramadass, Bala Haran, Parthasarathy M Gomadam, Ralph White, and Branko N
         Popov. Development of first principles capacity fade model for li-ion cells.
         Journal of the Electrochemical Society, 151(2):A196, 2004.
         doi:10.1149/1.1634273.
    * Xu2019 :
       - Shanshan Xu, Kuan-Hung Chen, Neil P Dasgupta, Jason B Siegel, and Anna G
         Stefanopoulou. Evolution of dead lithium growth in lithium metal batteries:
         experimentally validated model of the apparent capacity loss. Journal of The
         Electrochemical Society, 166(14):A3456, 2019.
"""

#
# Lithium-ion
#

NCA_Kim2011 = {
    "chemistry": "lithium_ion",
    "cell": "Kim2011",
    "negative electrode": "graphite_Kim2011",
    "separator": "separator_Kim2011",
    "positive electrode": "nca_Kim2011",
    "electrolyte": "lipf6_Kim2011",
    "experiment": "1C_discharge_from_full_Kim2011",
    "sei": "example",
    "citation": "Kim2011",
}

Ecker2015 = {
    "chemistry": "lithium_ion",
    "cell": "kokam_Ecker2015",
    "negative electrode": "graphite_Ecker2015",
    "separator": "separator_Ecker2015",
    "positive electrode": "LiNiCoO2_Ecker2015",
    "electrolyte": "lipf6_Ecker2015",
    "experiment": "1C_discharge_from_full_Ecker2015",
    "sei": "example",
    "citation": [
        "Ecker2015i",
        "Ecker2015ii",
        "Zhao2018",
        "Hales2019",
        "Richardson2020",
    ],
}

Marquis2019 = {
    "chemistry": "lithium_ion",
    "cell": "kokam_Marquis2019",
    "negative electrode": "graphite_mcmb2528_Marquis2019",
    "separator": "separator_Marquis2019",
    "positive electrode": "lico2_Marquis2019",
    "electrolyte": "lipf6_Marquis2019",
    "experiment": "1C_discharge_from_full_Marquis2019",
    "sei": "example",
    "citation": "Marquis2019",
}

Chen2020 = {
    "chemistry": "lithium_ion",
    "cell": "LGM50_Chen2020",
    "negative electrode": "graphite_Chen2020",
    "separator": "separator_Chen2020",
    "positive electrode": "nmc_Chen2020",
    "electrolyte": "lipf6_Nyman2008",
    "experiment": "1C_discharge_from_full_Chen2020",
    "sei": "example",
    "citation": "Chen2020",
}

Chen2020_plating = {
    "chemistry": "lithium_ion",
    "cell": "LGM50_Chen2020",
    "negative electrode": "graphite_Chen2020_plating",
    "separator": "separator_Chen2020",
    "positive electrode": "nmc_Chen2020",
    "electrolyte": "lipf6_Nyman2008",
    "experiment": "1C_discharge_from_full_Chen2020",
    "sei": "example",
    "lithium plating": "okane2020_Li_plating",
    "citation": "Chen2020",
}

Mohtat2020 = {
    "chemistry": "lithium_ion",
    "cell": "UMBL_Mohtat2020",
    "negative electrode": "graphite_UMBL_Mohtat2020",
    "separator": "separator_Mohtat2020",
    "positive electrode": "NMC_UMBL_Mohtat2020",
    "electrolyte": "LiPF6_Mohtat2020",
    "experiment": "1C_charge_from_empty_Mohtat2020",
    "sei": "example",
    "lithium plating": "yang2017_Li_plating",
    "citation": "Mohtat2020",
}

Ramadass2004 = {
    "chemistry": "lithium_ion",
    "cell": "sony_Ramadass2004",
    "negative electrode": "graphite_Ramadass2004",
    "separator": "separator_Ecker2015",  # no values found, relevance?
    "positive electrode": "lico2_Ramadass2004",
    "electrolyte": "lipf6_Ramadass2004",
    "experiment": "1C_discharge_from_full_Ramadass2004",
    "sei": "ramadass2004",
    "citation": "Ramadass2004",
}

Prada2013 = {
    "chemistry": "lithium_ion",
    "cell": "A123_Lain2019",
    "negative electrode": "graphite_Chen2020",
    "separator": "separator_Chen2020",
    "positive electrode": "LFP_Prada2013",
    "electrolyte": "lipf6_Nyman2008",
    "experiment": "4C_discharge_from_full_Prada2013",
    "citation": ["Chen2020", "Lain2019", "Prada2013"],
}

Ai2020 = {
    "chemistry": "lithium_ion",
    "cell": "Enertech_Ai2020",
    "negative electrode": "graphite_Ai2020",
    "separator": "separator_Ai2020",
    "positive electrode": "lico2_Ai2020",
    "electrolyte": "lipf6_Enertech_Ai2020",
    "experiment": "1C_discharge_from_full_Ai2020",
    "sei": "example",
    "citation": "Ai2019",
}

Xu2019 = {
    "chemistry": "lithium_ion",
    "cell": "li_metal_Xu2019",
    "negative electrode": "li_metal_Xu2019",
    "separator": "separator_Xu2019",
    "positive electrode": "NMC532_Xu2019",
    "electrolyte": "lipf6_Valoen2005",
    "experiment": "1C_discharge_from_full_Xu2019",
    "sei": "example",
    "citation": "Xu2019",
}

ORegan2021 = {
    "chemistry": "lithium_ion",
    "cell": "LGM50_ORegan2021",
    "negative electrode": "graphite_ORegan2021",
    "separator": "separator_ORegan2021",
    "positive electrode": "nmc_ORegan2021",
    "electrolyte": "lipf6_EC_EMC_3_7_Landesfeind2019",
    "experiment": "1C_discharge_from_full_ORegan2021",
    "citation": ["ORegan2021", "Chen2020"],
}

#
# Lead-acid
#

Sulzer2019 = {
    "chemistry": "lead_acid",
    "cell": "BBOXX_Sulzer2019",
    "negative electrode": "lead_Sulzer2019",
    "separator": "agm_Sulzer2019",
    "positive electrode": "lead_dioxide_Sulzer2019",
    "electrolyte": "sulfuric_acid_Sulzer2019",
    "experiment": "1C_discharge_from_full",
    "citation": "Sulzer2019physical",
}



#
# Solver class using Scipy's adaptive time stepper
#
import casadi
import pybamm

import scipy.integrate as it
import numpy as np


class ScipySolver(pybamm.BaseSolver):
    """Solve a discretised model, using scipy.integrate.solve_ivp.

    Parameters
    ----------
    method : str, optional
        The method to use in solve_ivp (default is "BDF")
    rtol : float, optional
        The relative tolerance for the solver (default is 1e-6).
    atol : float, optional
        The absolute tolerance for the solver (default is 1e-6).
    extrap_tol : float, optional
        The tolerance to assert whether extrapolation occurs or not (default is 0).
    extra_options : dict, optional
        Any options to pass to the solver.
        Please consult `SciPy documentation <https://tinyurl.com/yafgqg9y>`_ for
        details.
    """

    def __init__(
        self,
        method="BDF",
        rtol=1e-6,
        atol=1e-6,
        extrap_tol=0,
        extra_options=None,
    ):
        super().__init__(
            method=method,
            rtol=rtol,
            atol=atol,
            extrap_tol=extrap_tol,
        )
        self.ode_solver = True
        self.extra_options = extra_options or {}
        self.name = "Scipy solver ({})".format(method)
        pybamm.citations.register("Virtanen2020")

    def _integrate(self, model, t_eval, inputs_dict=None):
        """
        Solve a model defined by dydt with initial conditions y0.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate.
        t_eval : :class:`numpy.array`, size (k,)
            The times at which to compute the solution
        inputs_dict : dict, optional
            Any input parameters to pass to the model when solving

        Returns
        -------
        object
            An object containing the times and values of the solution, as well as
            various diagnostic messages.

        """
        # Save inputs dictionary, and if necessary convert inputs to a casadi vector
        inputs_dict = inputs_dict or {}
        if model.convert_to_format == "casadi":
            inputs = casadi.vertcat(*[x for x in inputs_dict.values()])
        else:
            inputs = inputs_dict

        extra_options = {**self.extra_options, "rtol": self.rtol, "atol": self.atol}

        # Initial conditions
        y0 = model.y0
        if isinstance(y0, casadi.DM):
            y0 = y0.full().flatten()

        # check for user-supplied Jacobian
        implicit_methods = ["Radau", "BDF", "LSODA"]
        if np.any([self.method in implicit_methods]):
            if model.jacobian_eval:
                extra_options.update(
                    {"jac": lambda t, y: model.jacobian_eval(t, y, inputs)}
                )

        # make events terminal so that the solver stops when they are reached
        if model.terminate_events_eval:

            def event_wrapper(event):
                def event_fn(t, y):
                    return event(t, y, inputs)

                event_fn.terminal = True
                return event_fn

            events = [event_wrapper(event) for event in model.terminate_events_eval]
            extra_options.update({"events": events})

        timer = pybamm.Timer()
        sol = it.solve_ivp(
            lambda t, y: model.rhs_eval(t, y, inputs),
            (t_eval[0], t_eval[-1]),
            y0,
            t_eval=t_eval,
            method=self.method,
            dense_output=True,
            **extra_options
        )
        integration_time = timer.time()

        if sol.success:
            # Set the reason for termination
            if sol.message == "A termination event occurred.":
                termination = "event"
                t_event = []
                for time in sol.t_events:
                    if time.size > 0:
                        t_event = np.append(t_event, np.max(time))
                t_event = np.array([np.max(t_event)])
                y_event = sol.sol(t_event)
            elif sol.message.startswith("The solver successfully reached the end"):
                termination = "final time"
                t_event = None
                y_event = np.array(None)
            sol = pybamm.Solution(
                sol.t, sol.y, model, inputs_dict, t_event, y_event, termination,
                sensitivities=bool(model.calculate_sensitivities)
            )
            sol.integration_time = integration_time
            return sol
        else:
            raise pybamm.SolverError(sol.message)



#
# Solver class using Scipy's adaptive time stepper
#
import casadi
import pybamm

import numpy as np
import importlib
import scipy.sparse as sparse

scikits_odes_spec = importlib.util.find_spec("scikits")
if scikits_odes_spec is not None:
    scikits_odes_spec = importlib.util.find_spec("scikits.odes")
    if scikits_odes_spec is not None:
        scikits_odes = importlib.util.module_from_spec(scikits_odes_spec)
        scikits_odes_spec.loader.exec_module(scikits_odes)


class ScikitsDaeSolver(pybamm.BaseSolver):
    """Solve a discretised model, using scikits.odes.

    Parameters
    ----------
    method : str, optional
        The method to use in solve_ivp (default is "BDF")
    rtol : float, optional
        The relative tolerance for the solver (default is 1e-6).
    atol : float, optional
        The absolute tolerance for the solver (default is 1e-6).
    root_method : str or pybamm algebraic solver class, optional
        The method to use to find initial conditions (for DAE solvers).
        If a solver class, must be an algebraic solver class.
        If "casadi",
        the solver uses casadi's Newton rootfinding algorithm to find initial
        conditions. Otherwise, the solver uses 'scipy.optimize.root' with method
        specified by 'root_method' (e.g. "lm", "hybr", ...)
    root_tol : float, optional
        The tolerance for the initial-condition solver (default is 1e-6).
    extrap_tol : float, optional
        The tolerance to assert whether extrapolation occurs or not (default is 0).
    extra_options : dict, optional
        Any options to pass to the solver.
        Please consult `scikits.odes documentation
        <https://bmcage.github.io/odes/dev/index.html>`_ for details.
        Some common keys:

        - 'max_steps': maximum (int) number of steps the solver can take
    """

    def __init__(
        self,
        method="ida",
        rtol=1e-6,
        atol=1e-6,
        root_method="casadi",
        root_tol=1e-6,
        extrap_tol=0,
        extra_options=None,
        max_steps="deprecated",
    ):
        if scikits_odes_spec is None:
            raise ImportError("scikits.odes is not installed")

        super().__init__(
            method, rtol, atol, root_method, root_tol, extrap_tol, max_steps
        )
        self.name = "Scikits DAE solver ({})".format(method)

        self.extra_options = extra_options or {}

        pybamm.citations.register("Malengier2018")
        pybamm.citations.register("Hindmarsh2000")
        pybamm.citations.register("Hindmarsh2005")

    def _integrate(self, model, t_eval, inputs_dict=None):
        """
        Solve a model defined by dydt with initial conditions y0.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate.
        t_eval : numeric type
            The times at which to compute the solution
        inputs_dict : dict, optional
            Any input parameters to pass to the model when solving

        """
        inputs_dict = inputs_dict or {}
        if model.convert_to_format == "casadi":
            inputs = casadi.vertcat(*[x for x in inputs_dict.values()])
        else:
            inputs = inputs_dict

        y0 = model.y0
        if isinstance(y0, casadi.DM):
            y0 = y0.full().flatten()

        residuals = model.residuals_eval
        events = model.terminate_events_eval
        jacobian = model.jacobian_eval
        mass_matrix = model.mass_matrix.entries

        def eqsres(t, y, ydot, return_residuals):
            return_residuals[:] = residuals(t, y, ydot, inputs)

        def rootfn(t, y, ydot, return_root):
            return_root[:] = [event(t, y, inputs) for event in events]

        extra_options = {
            **self.extra_options,
            "old_api": False,
            "rtol": self.rtol,
            "atol": self.atol,
        }

        if jacobian:
            jac_y0_t0 = jacobian(t_eval[0], y0, inputs)
            if sparse.issparse(jac_y0_t0):

                def jacfn(t, y, ydot, residuals, cj, J):
                    jac_eval = jacobian(t, y, inputs) - cj * mass_matrix
                    J[:][:] = jac_eval.toarray()

            else:

                def jacfn(t, y, ydot, residuals, cj, J):
                    jac_eval = jacobian(t, y, inputs) - cj * mass_matrix
                    J[:][:] = jac_eval

            extra_options.update({"jacfn": jacfn})

        if events:
            extra_options.update({"rootfn": rootfn, "nr_rootfns": len(events)})

        # solver works with ydot0 set to zero
        ydot0 = np.zeros_like(y0)

        # set up and solve
        dae_solver = scikits_odes.dae(self.method, eqsres, **extra_options)
        timer = pybamm.Timer()
        sol = dae_solver.solve(t_eval, y0, ydot0)
        integration_time = timer.time()

        # return solution, we need to tranpose y to match scipy's interface
        if sol.flag in [0, 2]:
            # 0 = solved for all t_eval
            if sol.flag == 0:
                termination = "final time"
            # 2 = found root(s)
            elif sol.flag == 2:
                termination = "event"
            if sol.roots.t is None:
                t_root = None
            else:
                t_root = sol.roots.t
            sol = pybamm.Solution(
                sol.values.t,
                np.transpose(sol.values.y),
                model,
                inputs_dict,
                t_root,
                np.transpose(sol.roots.y),
                termination,
            )
            sol.integration_time = integration_time
            return sol
        else:
            raise pybamm.SolverError(sol.message)



#
# Solver class using Scipy's adaptive time stepper
#
import casadi
import pybamm

import numpy as np
import importlib
import scipy.sparse as sparse

scikits_odes_spec = importlib.util.find_spec("scikits")
if scikits_odes_spec is not None:
    scikits_odes_spec = importlib.util.find_spec("scikits.odes")
    if scikits_odes_spec is not None:
        scikits_odes = importlib.util.module_from_spec(scikits_odes_spec)
        scikits_odes_spec.loader.exec_module(scikits_odes)


def have_scikits_odes():
    return scikits_odes_spec is not None


class ScikitsOdeSolver(pybamm.BaseSolver):
    """Solve a discretised model, using scikits.odes.

    Parameters
    ----------
    method : str, optional
        The method to use in solve_ivp (default is "BDF")
    rtol : float, optional
        The relative tolerance for the solver (default is 1e-6).
    atol : float, optional
        The absolute tolerance for the solver (default is 1e-6).
    extrap_tol : float, optional
        The tolerance to assert whether extrapolation occurs or not (default is 0).
    extra_options : dict, optional
        Any options to pass to the solver.
        Please consult `scikits.odes documentation
        <https://bmcage.github.io/odes/dev/index.html>`_ for details.
        Some common keys:

        - 'linsolver': can be 'dense' (= default), 'lapackdense', 'spgmr', 'spbcgs', \
        'sptfqmr'
    """

    def __init__(
        self,
        method="cvode",
        rtol=1e-6,
        atol=1e-6,
        extrap_tol=0,
        linsolver="deprecated",
        extra_options=None,
    ):
        if scikits_odes_spec is None:
            raise ImportError("scikits.odes is not installed")

        super().__init__(method, rtol, atol, extrap_tol=extrap_tol)
        self.extra_options = extra_options or {}
        if linsolver != "deprecated":
            raise ValueError(
                "linsolver has been deprecated. Pass 'linsolver' to extra_options "
                "dictionary instead"
            )
        self.ode_solver = True
        self.name = "Scikits ODE solver ({})".format(method)

        pybamm.citations.register("Malengier2018")
        pybamm.citations.register("Hindmarsh2000")
        pybamm.citations.register("Hindmarsh2005")

    def _integrate(self, model, t_eval, inputs_dict=None):
        """
        Solve a model defined by dydt with initial conditions y0.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate.
        t_eval : numeric type
            The times at which to compute the solution
        inputs_dict : dict, optional
            Any input parameters to pass to the model when solving

        """
        inputs_dict = inputs_dict or {}
        if model.rhs_eval.form == "casadi":
            inputs = casadi.vertcat(*[x for x in inputs_dict.values()])
        else:
            inputs = inputs_dict

        y0 = model.y0
        if isinstance(y0, casadi.DM):
            y0 = y0.full().flatten()

        derivs = model.rhs_eval
        events = model.terminate_events_eval
        jacobian = model.jacobian_eval

        def eqsydot(t, y, return_ydot):
            return_ydot[:] = derivs(t, y, inputs)

        def rootfn(t, y, return_root):
            return_root[:] = [event(t, y, inputs) for event in events]

        if jacobian:
            jac_y0_t0 = jacobian(t_eval[0], y0, inputs)
            if sparse.issparse(jac_y0_t0):

                def jacfn(t, y, fy, J):
                    J[:][:] = jacobian(t, y, inputs).toarray()

                def jac_times_vecfn(v, Jv, t, y, userdata):
                    Jv[:] = userdata._jac_eval * v
                    return 0

            else:

                def jacfn(t, y, fy, J):
                    J[:][:] = jacobian(t, y, inputs)

                def jac_times_vecfn(v, Jv, t, y, userdata):
                    Jv[:] = np.matmul(userdata._jac_eval, v)
                    return 0

            def jac_times_setupfn(t, y, fy, userdata):
                userdata._jac_eval = jacobian(t, y, inputs)
                return 0

        extra_options = {
            **self.extra_options,
            "old_api": False,
            "rtol": self.rtol,
            "atol": self.atol,
        }

        # Read linsolver (defaults to dense)
        linsolver = extra_options.get("linsolver", "dense")

        if jacobian:
            if linsolver in ("dense", "lapackdense"):
                extra_options.update({"jacfn": jacfn})
            elif linsolver in ("spgmr", "spbcgs", "sptfqmr"):
                extra_options.update(
                    {
                        "jac_times_setupfn": jac_times_setupfn,
                        "jac_times_vecfn": jac_times_vecfn,
                        "user_data": self,
                    }
                )

        if events:
            extra_options.update({"rootfn": rootfn, "nr_rootfns": len(events)})

        ode_solver = scikits_odes.ode(self.method, eqsydot, **extra_options)
        timer = pybamm.Timer()
        sol = ode_solver.solve(t_eval, y0)
        integration_time = timer.time()

        # return solution, we need to tranpose y to match scipy's ivp interface
        if sol.flag in [0, 2]:
            # 0 = solved for all t_eval
            if sol.flag == 0:
                termination = "final time"
            # 2 = found root(s)
            elif sol.flag == 2:
                termination = "event"
            if sol.roots.t is None:
                t_root = None
            else:
                t_root = sol.roots.t
            sol = pybamm.Solution(
                sol.values.t,
                np.transpose(sol.values.y),
                model,
                inputs_dict,
                t_root,
                np.transpose(sol.roots.y),
                termination,
            )
            sol.integration_time = integration_time
            return sol
        else:
            raise pybamm.SolverError(sol.message)



#
# SymPy overrides
#
import re

from sympy.printing.latex import LatexPrinter


class CustomPrint(LatexPrinter):
    """Override SymPy methods to match PyBaMM's requirements"""

    def _print_Derivative(self, expr):
        """Override :meth:`sympy.printing.latex.LatexPrinter._print_Derivative`"""
        eqn = super()._print_Derivative(expr)

        if getattr(expr, "force_partial", False) and "partial" not in eqn:
            var1, var2 = re.findall(r"^\\frac{(\w+)}{(\w+) .+", eqn)[0]
            eqn = eqn.replace(var1, "\partial").replace(var2, "\partial")

        return eqn


def custom_print_func(expr, **settings):
    return CustomPrint().doprint(expr)



#
# Basic Single Particle Model (SPM)
#
import pybamm
from .base_lithium_ion_model import BaseModel


class BasicSPM(BaseModel):
    """Single Particle Model (SPM) model of a lithium-ion battery, from [2]_.

    This class differs from the :class:`pybamm.lithium_ion.SPM` model class in that it
    shows the whole model in a single class. This comes at the cost of flexibility in
    combining different physical effects, and in general the main SPM class should be
    used instead.

    Parameters
    ----------
    name : str, optional
        The name of the model.

    References
    ----------
    .. [2] SG Marquis, V Sulzer, R Timms, CP Please and SJ Chapman. “An asymptotic
           derivation of a single particle model with electrolyte”. Journal of The
           Electrochemical Society, 166(15):A3693–A3706, 2019

    **Extends:** :class:`pybamm.lithium_ion.BaseModel`
    """

    def __init__(self, name="Single Particle Model"):
        super().__init__({}, name)
        pybamm.citations.register("Marquis2019")
        # `param` is a class containing all the relevant parameters and functions for
        # this model. These are purely symbolic at this stage, and will be set by the
        # `ParameterValues` class when the model is processed.
        param = self.param

        ######################
        # Variables
        ######################
        # Variables that depend on time only are created without a domain
        Q = pybamm.Variable("Discharge capacity [A.h]")
        # Variables that vary spatially are created with a domain
        c_s_n = pybamm.Variable(
            "X-averaged negative particle concentration", domain="negative particle"
        )
        c_s_p = pybamm.Variable(
            "X-averaged positive particle concentration", domain="positive particle"
        )

        # Constant temperature
        T = param.T_init

        ######################
        # Other set-up
        ######################

        # Current density
        i_cell = param.current_with_time
        j_n = i_cell / param.l_n
        j_p = -i_cell / param.l_p

        ######################
        # State of Charge
        ######################
        I = param.dimensional_current_with_time
        # The `rhs` dictionary contains differential equations, with the key being the
        # variable in the d/dt
        self.rhs[Q] = I * param.timescale / 3600
        # Initial conditions must be provided for the ODEs
        self.initial_conditions[Q] = pybamm.Scalar(0)

        ######################
        # Particles
        ######################

        # The div and grad operators will be converted to the appropriate matrix
        # multiplication at the discretisation stage
        N_s_n = -param.D_n(c_s_n, T) * pybamm.grad(c_s_n)
        N_s_p = -param.D_p(c_s_p, T) * pybamm.grad(c_s_p)
        self.rhs[c_s_n] = -(1 / param.C_n) * pybamm.div(N_s_n)
        self.rhs[c_s_p] = -(1 / param.C_p) * pybamm.div(N_s_p)
        # Surf takes the surface value of a variable, i.e. its boundary value on the
        # right side. This is also accessible via `boundary_value(x, "right")`, with
        # "left" providing the boundary value of the left side
        c_s_surf_n = pybamm.surf(c_s_n)
        c_s_surf_p = pybamm.surf(c_s_p)
        # Boundary conditions must be provided for equations with spatial derivatives
        self.boundary_conditions[c_s_n] = {
            "left": (pybamm.Scalar(0), "Neumann"),
            "right": (
                -param.C_n * j_n / param.a_R_n / param.D_n(c_s_surf_n, T),
                "Neumann",
            ),
        }
        self.boundary_conditions[c_s_p] = {
            "left": (pybamm.Scalar(0), "Neumann"),
            "right": (
                -param.C_p
                * j_p
                / param.a_R_p
                / param.gamma_p
                / param.D_p(c_s_surf_p, T),
                "Neumann",
            ),
        }
        # c_n_init and c_p_init are functions, but for the SPM we evaluate them at x=0
        # and x=1 since there is no x-dependence in the particles
        self.initial_conditions[c_s_n] = param.c_n_init(0)
        self.initial_conditions[c_s_p] = param.c_p_init(1)
        # Events specify points at which a solution should terminate
        self.events += [
            pybamm.Event(
                "Minimum negative particle surface concentration",
                pybamm.min(c_s_surf_n) - 0.01,
            ),
            pybamm.Event(
                "Maximum negative particle surface concentration",
                (1 - 0.01) - pybamm.max(c_s_surf_n),
            ),
            pybamm.Event(
                "Minimum positive particle surface concentration",
                pybamm.min(c_s_surf_p) - 0.01,
            ),
            pybamm.Event(
                "Maximum positive particle surface concentration",
                (1 - 0.01) - pybamm.max(c_s_surf_p),
            ),
        ]

        # Note that the SPM does not have any algebraic equations, so the `algebraic`
        # dictionary remains empty

        ######################
        # (Some) variables
        ######################
        # Interfacial reactions
        j0_n = param.j0_n(1, c_s_surf_n, T) / param.C_r_n
        j0_p = param.gamma_p * param.j0_p(1, c_s_surf_p, T) / param.C_r_p
        eta_n = (2 / param.ne_n) * pybamm.arcsinh(j_n / (2 * j0_n))
        eta_p = (2 / param.ne_p) * pybamm.arcsinh(j_p / (2 * j0_p))
        phi_s_n = 0
        phi_e = -eta_n - param.U_n(c_s_surf_n, T)
        phi_s_p = eta_p + phi_e + param.U_p(c_s_surf_p, T)
        V = phi_s_p

        whole_cell = ["negative electrode", "separator", "positive electrode"]
        # The `variables` dictionary contains all variables that might be useful for
        # visualising the solution of the model
        # Primary broadcasts are used to broadcast scalar quantities across a domain
        # into a vector of the right shape, for multiplying with other vectors
        self.variables = {
            "Negative particle surface concentration": pybamm.PrimaryBroadcast(
                c_s_surf_n, "negative electrode"
            ),
            "Electrolyte concentration": pybamm.PrimaryBroadcast(1, whole_cell),
            "Positive particle surface concentration": pybamm.PrimaryBroadcast(
                c_s_surf_p, "positive electrode"
            ),
            "Current [A]": I,
            "Negative electrode potential": pybamm.PrimaryBroadcast(
                phi_s_n, "negative electrode"
            ),
            "Electrolyte potential": pybamm.PrimaryBroadcast(phi_e, whole_cell),
            "Positive electrode potential": pybamm.PrimaryBroadcast(
                phi_s_p, "positive electrode"
            ),
            "Terminal voltage": V,
        }
        self.events += [
            pybamm.Event("Minimum voltage", V - param.voltage_low_cut),
            pybamm.Event("Maximum voltage", V - param.voltage_high_cut),
        ]

    def new_empty_copy(self):
        return pybamm.BaseModel.new_empty_copy(self)



#
# Base parameters class
#
import pybamm


class BaseParameters:
    """
    Overload the `__setattr__` method to record what the variable was called.
    """

    def __setattr__(self, name, value):
        if isinstance(value, pybamm.Symbol):
            value.print_name = name
        super().__setattr__(name, value)



from pybamm import exp, constants


def graphite_LGM50_diffusivity_Chen2020(sto, T):
    """
    LG M50 Graphite diffusivity as a function of stochiometry, in this case the
    diffusivity is taken to be a constant. The value is taken from [1].

    References
    ----------
    .. [1] Chang-Hui Chen, Ferran Brosa Planella, Kieran O’Regan, Dominika Gastol, W.
    Dhammika Widanage, and Emma Kendrick. "Development of Experimental Techniques for
    Parameterization of Multi-scale Lithium-ion Battery Models." Journal of the
    Electrochemical Society 167 (2020): 080534.

    Parameters
    ----------
    sto: :class:`pybamm.Symbol`
       Electrode stochiometry
    T: :class:`pybamm.Symbol`
       Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
       Solid diffusivity
    """

    D_ref = 3.3e-14
    E_D_s = 0  # to be implemented
    arrhenius = exp(E_D_s / constants.R * (1 / 298.15 - 1 / T))

    return D_ref * arrhenius



from pybamm import constants, Parameter


def plating_exchange_current_density_OKane2020(c_e, c_Li, T):
    """
    Exchange-current density for Li plating reaction [A.m-2].
    References
    ----------
    .. [1] O’Kane, Simon EJ, Ian D. Campbell, Mohamed WJ Marzook, Gregory J. Offer, and
    Monica Marinescu. "Physical origin of the differential voltage minimum associated
    with lithium plating in Li-ion batteries." Journal of The Electrochemical Society
    167, no. 9 (2020): 090540.
    Parameters
    ----------
    c_e : :class:`pybamm.Symbol`
        Electrolyte concentration [mol.m-3]
    c_Li : :class:`pybamm.Symbol`
        Plated lithium concentration [mol.m-3]
    T : :class:`pybamm.Symbol`
        Temperature [K]
    Returns
    -------
    :class:`pybamm.Symbol`
        Exchange-current density [A.m-2]
    """

    k_plating = Parameter("Lithium plating kinetic rate constant [m.s-1]")

    return constants.F * k_plating * c_e



#
# Base class for convection submodels
#
import pybamm


class BaseModel(pybamm.BaseSubModel):
    """Base class for convection submodels.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.

    **Extends:** :class:`pybamm.BaseSubModel`
    """

    def __init__(self, param, options=None):
        super().__init__(param, options=options)

    def _get_standard_whole_cell_velocity_variables(self, variables):
        """
        A private function to obtain the standard variables which
        can be derived from the fluid velocity.

        Parameters
        ----------
        variables : dict
            The existing variables in the model

        Returns
        -------
        variables : dict
            The variables which can be derived from the volume-averaged
            velocity.
        """

        vel_scale = self.param.velocity_scale

        if self.half_cell:
            v_box_n = None
        else:
            v_box_n = variables["Negative electrode volume-averaged velocity"]
        v_box_s = variables["Separator volume-averaged velocity"]
        v_box_p = variables["Positive electrode volume-averaged velocity"]

        v_box = pybamm.concatenation(v_box_n, v_box_s, v_box_p)

        variables = {
            "Volume-averaged velocity": v_box,
            "Volume-averaged velocity [m.s-1]": vel_scale * v_box,
        }

        return variables

    def _get_standard_whole_cell_acceleration_variables(self, variables):
        """
        A private function to obtain the standard variables which
        can be derived from the fluid velocity.

        Parameters
        ----------
        variables : dict
            The existing variables in the model

        Returns
        -------
        variables : dict
            The variables which can be derived from the volume-averaged
            velocity.
        """

        acc_scale = self.param.velocity_scale / self.param.L_x

        if self.half_cell:
            div_v_box_n = None
        else:
            div_v_box_n = variables["Negative electrode volume-averaged acceleration"]
        div_v_box_s = variables["Separator volume-averaged acceleration"]
        div_v_box_p = variables["Positive electrode volume-averaged acceleration"]

        div_v_box = pybamm.concatenation(div_v_box_n, div_v_box_s, div_v_box_p)
        div_v_box_av = pybamm.x_average(div_v_box)

        variables = {
            "Volume-averaged acceleration": div_v_box,
            "X-averaged volume-averaged acceleration": div_v_box_av,
            "Volume-averaged acceleration [m.s-1]": acc_scale * div_v_box,
            "X-averaged volume-averaged acceleration [m.s-1]": acc_scale * div_v_box_av,
        }

        return variables

    def _get_standard_whole_cell_pressure_variables(self, variables):
        """
        A private function to obtain the standard variables which
        can be derived from the pressure in the fluid.

        Parameters
        ----------
        variables : dict
            The existing variables in the model

        Returns
        -------
        variables : dict
            The variables which can be derived from the pressure.
        """
        if self.half_cell:
            p_n = None
        else:
            p_n = variables["Negative electrode pressure"]
        p_s = variables["Separator pressure"]
        p_p = variables["Positive electrode pressure"]

        p = pybamm.concatenation(p_n, p_s, p_p)

        variables = {"Pressure": p}

        return variables



#
# Bulter volmer class
#

import pybamm
from .base_kinetics import BaseKinetics


class SymmetricButlerVolmer(BaseKinetics):
    """
    Submodel which implements the symmetric forward Butler-Volmer equation:

    .. math::
        j = 2 * j_0(c) * \\sinh( (ne / (2 * (1 + \\Theta T)) * \\eta_r(c))

    Parameters
    ----------
    param : parameter class
        model parameters
    domain : str
        The domain to implement the model, either: 'Negative' or 'Positive'.
    reaction : str
        The name of the reaction being implemented
    options: dict
        A dictionary of options to be passed to the model.
        See :class:`pybamm.BaseBatteryModel`

    **Extends:** :class:`pybamm.interface.kinetics.BaseKinetics`
    """

    def __init__(self, param, domain, reaction, options):
        super().__init__(param, domain, reaction, options)

    def _get_kinetics(self, j0, ne, eta_r, T, u):
        prefactor = ne / (2 * (1 + self.param.Theta * T))
        return 2 * u * j0 * pybamm.sinh(prefactor * eta_r)

    def _get_dj_dc(self, variables):
        """See :meth:`pybamm.interface.kinetics.BaseKinetics._get_dj_dc`"""
        (
            c_e,
            delta_phi,
            j0,
            ne,
            ocp,
            T,
            u,
        ) = self._get_interface_variables_for_first_order(variables)
        eta_r = delta_phi - ocp
        prefactor = ne / (2 * (1 + self.param.Theta * T))
        return (2 * u * j0.diff(c_e) * pybamm.sinh(prefactor * eta_r)) - (
            2 * u * j0 * prefactor * ocp.diff(c_e) * pybamm.cosh(prefactor * eta_r)
        )

    def _get_dj_ddeltaphi(self, variables):
        """See :meth:`pybamm.interface.kinetics.BaseKinetics._get_dj_ddeltaphi`"""
        _, delta_phi, j0, ne, ocp, T, u = self._get_interface_variables_for_first_order(
            variables
        )
        eta_r = delta_phi - ocp
        prefactor = ne / (2 * (1 + self.param.Theta * T))
        return 2 * u * j0 * prefactor * pybamm.cosh(prefactor * eta_r)


class AsymmetricButlerVolmer(BaseKinetics):
    """
    Submodel which implements the asymmetric forward Butler-Volmer equation

    Parameters
    ----------
    param : parameter class
        model parameters
    domain : str
        The domain to implement the model, either: 'Negative' or 'Positive'.
    reaction : str
        The name of the reaction being implemented
    options: dict
        A dictionary of options to be passed to the model.
        See :class:`pybamm.BaseBatteryModel`

    **Extends:** :class:`pybamm.interface.kinetics.BaseKinetics`
    """

    def __init__(self, param, domain, reaction, options):
        super().__init__(param, domain, reaction, options)

    def _get_kinetics(self, j0, ne, eta_r, T, u):
        if self.domain == "Negative":
            alpha = self.param.alpha_bv_n
        elif self.domain == "Positive":
            alpha = self.param.alpha_bv_p
        arg_ox = ne * alpha * eta_r / (1 + self.param.Theta * T)
        arg_red = -ne * (1 - alpha) * eta_r / (1 + self.param.Theta * T)
        return u * j0 * (pybamm.exp(arg_ox) - pybamm.exp(arg_red))



from pybamm import exp, constants


def graphite_diffusivity_Ecker2015(sto, T):
    """
    Graphite diffusivity as a function of stochiometry [1, 2, 3].

    References
    ----------
     .. [1] Ecker, Madeleine, et al. "Parameterization of a physico-chemical model of
    a lithium-ion battery i. determination of parameters." Journal of the
    Electrochemical Society 162.9 (2015): A1836-A1848.
    .. [2] Ecker, Madeleine, et al. "Parameterization of a physico-chemical model of
    a lithium-ion battery ii. model validation." Journal of The Electrochemical
    Society 162.9 (2015): A1849-A1857.
    .. [3] Richardson, Giles, et. al. "Generalised single particle models for
    high-rate operation of graded lithium-ion electrodes: Systematic derivation
    and validation." Electrochemica Acta 339 (2020): 135862

    Parameters
    ----------
    sto: :class:`pybamm.Symbol`
        Electrode stochiometry
    T: :class:`pybamm.Symbol`
        Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
        Solid diffusivity
    """

    D_ref = 8.4e-13 * exp(-11.3 * sto) + 8.2e-15
    E_D_s = 3.03e4
    arrhenius = exp(-E_D_s / (constants.R * T)) * exp(E_D_s / (constants.R * 296))

    return D_ref * arrhenius



#
# Subodels for a lithium metal electrode
#
import pybamm
from .base_ohm import BaseModel


class LithiumMetalBaseModel(BaseModel):
    def __init__(self, param, options=None):
        super().__init__(param, "Negative", options=options)

    def _get_li_metal_interface_variables(self, delta_phi_s, phi_s, phi_e):
        param = self.param
        pot_scale = param.potential_scale
        delta_phi_s_dim = pot_scale * delta_phi_s

        variables = {
            "Negative electrode potential drop": delta_phi_s,
            "Negative electrode potential drop [V]": delta_phi_s_dim,
            "X-averaged negative electrode ohmic losses": delta_phi_s / 2,
            "X-averaged negative electrode ohmic losses [V]": delta_phi_s_dim / 2,
            "Lithium metal interface electrode potential": phi_s,
            "Lithium metal interface electrode potential [V]": pot_scale * phi_s,
            "Lithium metal interface electrolyte potential": phi_e,
            "Lithium metal interface electrolyte potential [V]": param.U_n_ref
            + pot_scale * phi_e,
        }
        return variables


class LithiumMetalSurfaceForm(LithiumMetalBaseModel):
    """Explicit model for potential drop across a lithium metal electrode.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.

    **Extends:** :class:`pybamm.electrode.li_metal.LithiumMetalBaseModel`
    """

    def get_fundamental_variables(self):
        ocp_ref = self.param.U_n_ref
        pot_scale = self.param.potential_scale

        delta_phi = pybamm.Variable(
            "Lithium metal interface surface potential difference",
            domain="current collector",
        )
        variables = {
            "Lithium metal interface surface potential difference": delta_phi,
            "Lithium metal interface surface potential difference [V]": ocp_ref
            + delta_phi * pot_scale,
        }

        return variables

    def get_coupled_variables(self, variables):
        param = self.param

        i_boundary_cc = variables["Current collector current density"]
        T_n = variables["Negative current collector temperature"]
        l_n = param.l_n
        delta_phi_s = i_boundary_cc * l_n / param.sigma_n(T_n)

        phi_s_cn = variables["Negative current collector potential"]
        delta_phi = variables["Lithium metal interface surface potential difference"]

        # Potentials at the anode/separator interface
        phi_s = phi_s_cn - delta_phi_s
        phi_e = phi_s - delta_phi

        variables.update(
            self._get_li_metal_interface_variables(delta_phi_s, phi_s, phi_e)
        )
        return variables

    def set_initial_conditions(self, variables):
        delta_phi = variables["Lithium metal interface surface potential difference"]
        delta_phi_init = self.param.U_n_init

        self.initial_conditions = {delta_phi: delta_phi_init}

    def set_algebraic(self, variables):
        j_pl = variables["Lithium metal plating current density"]
        j_sei = variables["SEI interfacial current density"]
        sum_j = j_pl + j_sei

        i_cc = variables["Current collector current density"]
        delta_phi = variables["Lithium metal interface surface potential difference"]

        self.algebraic[delta_phi] = i_cc - sum_j


class LithiumMetalExplicit(LithiumMetalBaseModel):
    """Explicit model for potential drop across a lithium metal electrode.

    Parameters
    ----------
    param : parameteslackr class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.

    **Extends:** :class:`pybamm.electrode.li_metal.LithiumMetalBaseModel`
    """

    def get_coupled_variables(self, variables):
        param = self.param

        i_boundary_cc = variables["Current collector current density"]
        T_n = variables["Negative current collector temperature"]
        l_n = param.l_n
        delta_phi_s = i_boundary_cc * l_n / param.sigma_n(T_n)

        phi_s_cn = variables["Negative current collector potential"]
        delta_phi = variables["Lithium metal interface surface potential difference"]

        phi_s = phi_s_cn - delta_phi_s
        phi_e = phi_s - delta_phi

        variables.update(
            self._get_li_metal_interface_variables(delta_phi_s, phi_s, phi_e)
        )
        return variables



#
# Lithium-ion base model class
#
import pybamm


class BaseModel(pybamm.BaseBatteryModel):
    """
    Overwrites default parameters from Base Model with default parameters for
    lithium-ion models

    **Extends:** :class:`pybamm.BaseBatteryModel`

    """

    def __init__(self, options=None, name="Unnamed lithium-ion model", build=False):
        super().__init__(options, name)
        self.param = pybamm.LithiumIonParameters(options)

        # Assess whether the submodel is a half-cell model
        self.half_cell = self.options["working electrode"] != "both"

        # Default timescale is discharge timescale
        self.timescale = self.param.tau_discharge

        # Set default length scales
        self.length_scales = {
            "negative electrode": self.param.L_x,
            "separator": self.param.L_x,
            "positive electrode": self.param.L_x,
            "positive particle": self.param.R_p_typ,
            "positive particle size": self.param.R_p_typ,
            "current collector y": self.param.L_z,
            "current collector z": self.param.L_z,
        }

        # Add negative particle domains only if not a half cell model
        if not self.half_cell:
            self.length_scales.update(
                {
                    "negative particle": self.param.R_n_typ,
                    "negative particle size": self.param.R_n_typ,
                }
            )
        self.set_standard_output_variables()

    @property
    def default_parameter_values(self):
        if self.half_cell:
            return pybamm.ParameterValues("Xu2019")
        else:
            return pybamm.ParameterValues("Marquis2019")

    @property
    def default_quick_plot_variables(self):
        if self.half_cell:
            return [
                "Electrolyte concentration [mol.m-3]",
                "Positive particle surface concentration [mol.m-3]",
                "Current [A]",
                "Electrolyte potential [V]",
                "Positive electrode potential [V]",
                "Terminal voltage [V]",
            ]
        else:
            return [
                "Negative particle surface concentration [mol.m-3]",
                "Electrolyte concentration [mol.m-3]",
                "Positive particle surface concentration [mol.m-3]",
                "Current [A]",
                "Negative electrode potential [V]",
                "Electrolyte potential [V]",
                "Positive electrode potential [V]",
                "Terminal voltage [V]",
            ]

    def set_standard_output_variables(self):
        super().set_standard_output_variables()

        # Particle concentration position
        var = pybamm.standard_spatial_vars
        self.variables.update({"r_p": var.r_p, "r_p [m]": var.r_p * self.param.R_p_typ})
        if not self.half_cell:
            self.variables.update(
                {"r_n": var.r_n, "r_n [m]": var.r_n * self.param.R_n_typ}
            )

    def set_degradation_variables(self):
        """Sets variables that quantify degradation (LAM, LLI, etc)"""
        param = self.param

        # LAM
        if self.half_cell:
            n_Li_n = pybamm.Scalar(0)
            LAM_ne = pybamm.Scalar(0)
        else:
            C_n = self.variables["Negative electrode capacity [A.h]"]
            n_Li_n = self.variables["Total lithium in negative electrode [mol]"]
            LAM_ne = (1 - C_n / param.C_n_init) * 100

        C_p = self.variables["Positive electrode capacity [A.h]"]

        LAM_pe = (1 - C_p / param.C_p_init) * 100

        # LLI
        n_Li_e = self.variables["Total lithium in electrolyte [mol]"]
        n_Li_p = self.variables["Total lithium in positive electrode [mol]"]
        n_Li_particles = n_Li_n + n_Li_p
        n_Li = n_Li_particles + n_Li_e

        # LLI is usually defined based only on the percentage lithium lost from
        # particles
        LLI = (1 - n_Li_particles / param.n_Li_particles_init) * 100
        LLI_tot = (1 - n_Li / param.n_Li_init) * 100

        self.variables.update(
            {
                "LAM_ne [%]": LAM_ne,
                "LAM_pe [%]": LAM_pe,
                "LLI [%]": LLI,
                "Loss of active material in negative electrode [%]": LAM_ne,
                "Loss of active material in positive electrode [%]": LAM_pe,
                "Loss of lithium inventory [%]": LLI,
                "Loss of lithium inventory, including electrolyte [%]": LLI_tot,
                # Total lithium
                "Total lithium [mol]": n_Li,
                "Total lithium in particles [mol]": n_Li_particles,
                # Lithium lost
                "Total lithium lost [mol]": param.n_Li_init - n_Li,
                "Total lithium lost from particles [mol]": param.n_Li_particles_init
                - n_Li_particles,
                "Total lithium lost from electrolyte [mol]": param.n_Li_e_init - n_Li_e,
            }
        )

        # Lithium lost to side reactions
        # Different way of measuring LLI but should give same value
        LLI_sei = self.variables["Loss of lithium to SEI [mol]"]
        if self.half_cell:
            LLI_pl = pybamm.Scalar(0)
        else:
            LLI_pl = self.variables["Loss of lithium to lithium plating [mol]"]

        LLI_reactions = LLI_sei + LLI_pl
        self.variables.update(
            {
                "Total lithium lost to side reactions [mol]": LLI_reactions,
                "Total capacity lost to side reactions [A.h]": LLI_reactions
                * param.F
                / 3600,
            }
        )

    def set_summary_variables(self):
        """
        Sets the default summary variables.
        """
        summary_variables = [
            "Positive electrode capacity [A.h]",
            # LAM, LLI
            "Loss of active material in positive electrode [%]",
            "Loss of lithium inventory [%]",
            "Loss of lithium inventory, including electrolyte [%]",
            # Total lithium
            "Total lithium [mol]",
            "Total lithium in electrolyte [mol]",
            "Total lithium in positive electrode [mol]",
            "Total lithium in particles [mol]",
            # Lithium lost
            "Total lithium lost [mol]",
            "Total lithium lost from particles [mol]",
            "Total lithium lost from electrolyte [mol]",
            "Loss of lithium to SEI [mol]",
            "Loss of capacity to SEI [A.h]",
            "Total lithium lost to side reactions [mol]",
            "Total capacity lost to side reactions [A.h]",
            # Resistance
            "Local ECM resistance [Ohm]",
        ]

        if not self.half_cell:
            summary_variables += [
                "Negative electrode capacity [A.h]",
                "Loss of active material in negative electrode [%]",
                "Total lithium in negative electrode [mol]",
                "Loss of lithium to lithium plating [mol]",
                "Loss of capacity to lithium plating [A.h]",
            ]

        self.summary_variables = summary_variables

    def set_sei_submodel(self):
        if self.half_cell:
            reaction_loc = "interface"
        elif self.x_average:
            reaction_loc = "x-average"
        else:
            reaction_loc = "full electrode"

        if self.options["SEI"] == "none":
            self.submodels["sei"] = pybamm.sei.NoSEI(self.param, self.options)
        elif self.options["SEI"] == "constant":
            self.submodels["sei"] = pybamm.sei.ConstantSEI(self.param, self.options)
        else:
            self.submodels["sei"] = pybamm.sei.SEIGrowth(
                self.param, reaction_loc, self.options
            )

    def set_lithium_plating_submodel(self):
        if self.options["lithium plating"] == "none":
            self.submodels["lithium plating"] = pybamm.lithium_plating.NoPlating(
                self.param, self.options
            )
        else:
            self.submodels["lithium plating"] = pybamm.lithium_plating.Plating(
                self.param, self.x_average, self.options
            )

    def set_other_reaction_submodels_to_zero(self):
        self.submodels["negative oxygen interface"] = pybamm.kinetics.NoReaction(
            self.param, "Negative", "lithium-ion oxygen"
        )
        self.submodels["positive oxygen interface"] = pybamm.kinetics.NoReaction(
            self.param, "Positive", "lithium-ion oxygen"
        )

    def set_crack_submodel(self):
        for domain in ["Negative", "Positive"]:
            crack = getattr(self.options, domain.lower())["particle mechanics"]
            if crack == "none":
                pass
            elif crack == "swelling only":
                self.submodels[
                    domain.lower() + " particle mechanics"
                ] = pybamm.particle_mechanics.SwellingOnly(self.param, domain)
            elif crack == "swelling and cracking":
                self.submodels[
                    domain.lower() + " particle mechanics"
                ] = pybamm.particle_mechanics.CrackPropagation(
                    self.param, domain, self.x_average
                )

    def set_active_material_submodel(self):
        for domain in ["Negative", "Positive"]:
            lam = getattr(self.options, domain.lower())["loss of active material"]
            if lam == "none":
                self.submodels[
                    domain.lower() + " active material"
                ] = pybamm.active_material.Constant(self.param, domain, self.options)
            else:
                self.submodels[
                    domain.lower() + " active material"
                ] = pybamm.active_material.LossActiveMaterial(
                    self.param, domain, self.options, self.x_average
                )

    def set_porosity_submodel(self):
        if (
            self.options["SEI porosity change"] == "false"
            and self.options["lithium plating porosity change"] == "false"
        ):
            self.submodels["porosity"] = pybamm.porosity.Constant(
                self.param, self.options
            )
        elif (
            self.options["SEI porosity change"] == "true"
            or self.options["lithium plating porosity change"] == "true"
        ):
            self.submodels["porosity"] = pybamm.porosity.ReactionDriven(
                self.param, self.options, self.x_average
            )

    def set_li_metal_counter_electrode_submodels(self):
        if (
            self.options["SEI"] in ["none", "constant"]
            and self.options["intercalation kinetics"] == "symmetric Butler-Volmer"
            and self.options["surface form"] == "false"
        ):
            # only symmetric Butler-Volmer can be inverted
            self.submodels[
                "counter electrode potential"
            ] = pybamm.electrode.ohm.LithiumMetalExplicit(self.param, self.options)
            self.submodels[
                "counter electrode interface"
            ] = pybamm.kinetics.InverseButlerVolmer(
                self.param, "Negative", "lithium metal plating", self.options
            )  # assuming symmetric reaction for now so we can take the inverse
            self.submodels[
                "counter electrode interface current"
            ] = pybamm.kinetics.CurrentForInverseButlerVolmerLithiumMetal(
                self.param, "Negative", "lithium metal plating", self.options
            )
        else:
            self.submodels[
                "counter electrode potential"
            ] = pybamm.electrode.ohm.LithiumMetalSurfaceForm(self.param, self.options)
            self.submodels["counter electrode interface"] = self.intercalation_kinetics(
                self.param, "Negative", "lithium metal plating", self.options
            )

        # For half-cell models, remove negative electrode submodels
        # that are not needed before building
        # We do this whether the working electrode is 'positive' or 'negative' since
        # the half-cell models are always defined assuming the positive electrode is
        # the working electrode

        # This should be done before `self.build_model`, which is the expensive part

        # Models added specifically for the counter electrode have been labelled with
        # "counter electrode" so as not to be caught by this check
        self.submodels = {
            k: v
            for k, v in self.submodels.items()
            if not (k.startswith("negative") or k == "lithium plating")
        }



#
# Class for particle-size distributions, one distribution at every
# x location of the electrode, with uniform concentration in each
# particle
#
import pybamm

from .base_distribution import BaseSizeDistribution


class UniformProfile(BaseSizeDistribution):
    """
    Class for molar conservation in particle-size distributions, one
    distribution at every x location of the electrode,
    with a uniform concentration within each particle (in r). Concentration varies
    with R (particle size), and x (electrode coordinate).

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain of the model either 'Negative' or 'Positive'


    **Extends:** :class:`pybamm.particle.size_distribution.BaseSizeDistribution`
    """

    def __init__(self, param, domain):
        super().__init__(param, domain)
        pybamm.citations.register("Kirk2021")

    def get_fundamental_variables(self):
        # The concentration is uniform throughout each particle, so we
        # can just use the surface value.

        if self.domain == "Negative":
            # distribution variables
            c_s_surf_distribution = pybamm.Variable(
                "Negative particle surface concentration distribution",
                domain="negative particle size",
                auxiliary_domains={
                    "secondary": "negative electrode",
                    "tertiary": "current collector",
                },
                bounds=(0, 1),
            )
            R = pybamm.standard_spatial_vars.R_n

        elif self.domain == "Positive":
            # distribution variables
            c_s_surf_distribution = pybamm.Variable(
                "Positive particle surface concentration distribution",
                domain="positive particle size",
                auxiliary_domains={
                    "secondary": "positive electrode",
                    "tertiary": "current collector",
                },
                bounds=(0, 1),
            )
            R = pybamm.standard_spatial_vars.R_p

        variables = self._get_distribution_variables(R)

        # Standard concentration distribution variables (size-dependent)
        variables.update(
            self._get_standard_concentration_distribution_variables(
                c_s_surf_distribution
            )
        )
        # Flux variables (size-dependent)
        variables.update(
            self._get_standard_flux_distribution_variables(pybamm.Scalar(0))
        )

        # Standard size-averaged variables. Average concentrations using
        # the volume-weighted distribution since they are volume-based
        # quantities. Necessary for output variables "Total lithium in
        # negative electrode [mol]", etc, to be calculated correctly
        f_v_dist = variables[
            self.domain + " volume-weighted particle-size distribution"
        ]
        c_s_surf = pybamm.Integral(f_v_dist * c_s_surf_distribution, R)
        c_s = pybamm.PrimaryBroadcast(
            c_s_surf, [self.domain.lower() + " particle"]
        )
        c_s_xav = pybamm.x_average(c_s)
        variables.update(self._get_standard_concentration_variables(c_s, c_s_xav))

        # Size-averaged flux variables
        N_s = pybamm.FullBroadcastToEdges(
            0,
            [self.domain.lower() + " particle"],
            auxiliary_domains={
                "secondary": self.domain.lower() + " electrode",
                "tertiary": "current collector",
            },
        )
        N_s_xav = pybamm.FullBroadcastToEdges(
            0, self.domain.lower() + " particle", "current collector"
        )
        variables.update(self._get_standard_flux_variables(N_s, N_s_xav))
        return variables

    def get_coupled_variables(self, variables):
        variables.update(self._get_total_concentration_variables(variables))
        return variables

    def set_rhs(self, variables):
        c_s_surf_distribution = variables[
            self.domain
            + " particle surface concentration distribution"
        ]
        j_distribution = variables[
            self.domain
            + " electrode interfacial current density distribution"
        ]
        R = variables[self.domain + " particle sizes"]

        if self.domain == "Negative":
            self.rhs = {
                c_s_surf_distribution: -3
                * j_distribution
                / self.param.a_R_n
                / R
            }
        elif self.domain == "Positive":
            self.rhs = {
                c_s_surf_distribution: -3
                * j_distribution
                / self.param.a_R_p
                / self.param.gamma_p
                / R
            }

    def set_initial_conditions(self, variables):
        c_s_surf_distribution = variables[
            self.domain
            + " particle surface concentration distribution"
        ]

        if self.domain == "Negative":
            # Broadcast x_n to particle size domain
            x_n = pybamm.PrimaryBroadcast(
                pybamm.standard_spatial_vars.x_n, "negative particle size"
            )
            c_init = self.param.c_n_init(x_n)

        elif self.domain == "Positive":
            # Broadcast x_p to particle size domain
            x_p = pybamm.PrimaryBroadcast(
                pybamm.standard_spatial_vars.x_p, "positive particle size"
            )
            c_init = self.param.c_p_init(x_p)

        self.initial_conditions = {c_s_surf_distribution: c_init}



from .base_fickian import BaseFickian
from .fickian_diffusion import FickianDiffusion
from .x_averaged_fickian_diffusion import XAveragedFickianDiffusion
from .x_averaged_polynomial_profile import XAveragedPolynomialProfile
from .polynomial_profile import PolynomialProfile
